@@ -5,6 +5,7 @@
 from pathlib import Path
 
 import numpy as np
+from scipy import stats
 from nilearn import plotting
 from tqdm import trange
 
@@ -124,8 +125,8 @@ def mean_coherence_from_spectra(
     -------
     coh : np.ndarray
         Mean coherence over a frequency band for each component of each mode.
-        Shape is (n_components, n_modes, n_channels, n_channels). Axis of
-        length 1 are squeezed.
+        Shape is (n_components, n_modes, n_channels, n_channels) or
+        (n_modes, n_channels, n_channels) or (n_channels, n_channels).
     """
 
     # Validation
@@ -165,9 +166,6 @@ def mean_coherence_from_spectra(
         # Concatenate over modes
         coh = coherence[i].reshape(-1, n_f)
 
-        # Same frequencies give nan coherences, replace with zero
-        coh = np.nan_to_num(coh)
-
         if components is not None:
             # Coherence for each spectral component
             coh = components @ coh.T
@@ -188,15 +186,34 @@ def mean_coherence_from_spectra(
     return np.squeeze(c)
 
 
+def mean_connections(conn_map):
+    """Average the edges for each node.
+
+    Parameters
+    ----------
+    conn_map : np.ndarray
+        A (..., n_channels, n_channels) connectivity matrix.
+
+    Returns
+    -------
+    mean_connections : np.ndarray
+        A (..., n_channels) matrix.
+    """
+    return np.mean(conn_map, axis=-1)
+
+
 def gmm_threshold(
     conn_map,
     subtract_mean=False,
+    mean_weights=None,
     standardize=False,
+    p_value=None,
+    keep_positive_only=False,
     one_component_percentile=0,
     n_sigma=0,
-    sklearn_kwargs=None,
+    sklearn_kwargs={},
     filename=None,
-    plot_kwargs=None,
+    plot_kwargs={},
 ):
     """Threshold a connectivity matrix using the GMM method.
 
@@ -206,11 +223,20 @@ def gmm_threshold(
     ----------
     conn_map : np.ndarray
         Connectivity matrix. Shape must be (n_components, n_modes, n_channels,
-        n_channels) or (n_modes, n_channels, n_channels) or (n_channels, n_channesl).
+        n_channels) or (n_modes, n_channels, n_channels) or (n_channels, n_channels).
     subtract_mean : bool
         Should we subtract the mean over modes before fitting a GMM?
+    mean_weights: np.ndarray
+        Numpy array with weightings for each mode/state to use to calculate the mean.
+        Default is equal weighting.
     standardize : bool
         Should we standardize the input to the GMM?
+    p_value : float
+        Used to determine a threshold. We ensure the data points assigned
+        to the 'on' component have a probability of less than p_value of
+        belonging to the 'off' component.
+    keep_positive_only : bool
+        Should we only keep positive values to fit a GMM to?
     one_component_percentile : float
         Percentile threshold if only one component is found.
         Should be a between 0 and 100. E.g. for the 95th percentile,
@@ -236,26 +262,32 @@ def gmm_threshold(
     percentile = fit_gmm(
         conn_map,
         subtract_mean,
+        mean_weights,
         standardize,
+        p_value,
+        keep_positive_only,
         one_component_percentile,
         n_sigma,
         sklearn_kwargs,
         filename,
         plot_kwargs,
     )
-    conn_map = threshold(conn_map, percentile, subtract_mean, return_edges=False)
+    conn_map = threshold(conn_map, percentile, subtract_mean)
     return conn_map
 
 
 def fit_gmm(
     conn_map,
     subtract_mean=False,
+    mean_weights=None,
     standardize=False,
+    p_value=None,
+    keep_positive_only=False,
     one_component_percentile=0,
     n_sigma=0,
-    sklearn_kwargs=None,
+    sklearn_kwargs={},
     filename=None,
-    plot_kwargs=None,
+    plot_kwargs={},
 ):
     """Fit a two component Gaussian mixture model to connections to identify a
     threshold.
@@ -266,8 +298,17 @@ def fit_gmm(
         Connectivity map.
     subtract_mean : bool
         Should we subtract the mean over modes before fitting a GMM?
+    mean_weights: np.ndarray
+        Numpy array with weightings for each mode/state to use to calculate the mean.
+        Default is equal weighting.
     standardize : bool
         Should we standardize the input to the GMM?
+    p_value : float
+        Used to determine a threshold. We ensure the data points assigned
+        to the 'on' component have a probability of less than p_value of
+        belonging to the 'off' component.
+    keep_positive_only : bool
+        Should we only keep positive values to fit a GMM to?
     one_component_percentile : float
         Percentile threshold if only one component is found.
         Should be a between 0 and 100. E.g. for the 95th percentile,
@@ -278,7 +319,8 @@ def fit_gmm(
         have two components.
     sklearn_kwargs : dict
         Dictionary of keyword arguments to pass to
-        sklearn.mixture.BayesianGaussianMixture().
+        sklearn.mixture.GaussianMixture(). Default is
+        {"max_iter": 5000, "n_init": 10}.
     filename : str
         Filename to save fit to.
     plot_kwargs : dict
@@ -299,8 +341,8 @@ def fit_gmm(
         + "or (n_channels, n_channels).",
     )
 
-    if sklearn_kwargs is None:
-        sklearn_kwargs = {"max_iter": 5000, "n_init": 10}
+    default_sklearn_kwargs = {"max_iter": 5000, "n_init": 10}
+    sklearn_kwargs = override_dict_defaults(default_sklearn_kwargs, sklearn_kwargs)
 
     # Number of components, modes and channels
     n_components = conn_map.shape[0]
@@ -308,7 +350,7 @@ def fit_gmm(
     n_channels = conn_map.shape[2]
 
     # Mean over modes
-    mean_conn_map = np.mean(conn_map, axis=1)
+    mean_conn_map = np.average(conn_map, axis=1, weights=mean_weights)
 
     # Indices for off diagonal elements
     m, n = np.triu_indices(n_channels, k=1)
@@ -324,6 +366,14 @@ def fit_gmm(
             else:
                 c = conn_map[i, j, m, n]
 
+            if keep_positive_only:
+                # Only keep positive entries
+                # (this is what's done in MATLAB OSL's teh_graph_gmm_fit)
+                c = c[c > 0]
+                if len(c) == 0:
+                    percentiles[i, j] = 100
+                    continue
+
             # Output filename
             if filename is not None:
                 plot_filename = (
@@ -339,18 +389,21 @@ def fit_gmm(
                 plot_filename = None
 
             # Fit a GMM to get class labels
-            percentiles[i, j] = gmm.fit_gaussian_mixture(
+            threshold = gmm.fit_gaussian_mixture(
                 c,
-                bayesian=True,
                 standardize=standardize,
-                sklearn_kwargs=sklearn_kwargs,
+                p_value=p_value,
                 one_component_percentile=one_component_percentile,
                 n_sigma=n_sigma,
+                sklearn_kwargs=sklearn_kwargs,
                 plot_filename=plot_filename,
                 plot_kwargs=plot_kwargs,
                 print_message=False,
             )
 
+            # Calculate the percentile from the threshold
+            percentiles[i, j] = stats.percentileofscore(c, threshold)
+
     return np.squeeze(percentiles)
 
 
