@@ -8,6 +8,7 @@ import numpy as np
 from scipy.signal.windows import dpss, hann
 from sklearn.decomposition import non_negative_factorization
 from tqdm import trange
+from pqdm.processes import pqdm
 
 from osl_dynamics import array_ops
 from osl_dynamics.analysis import regression, time_series
@@ -60,7 +61,8 @@ def coherence_spectra(power_spectra, print_message=True):
                     power_spectra[i, j, j].real * power_spectra[i, k, k].real
                 )
 
-    return coherences
+    # Zero nan values
+    return np.nan_to_num(coherences)
 
 
 def decompose_spectra(
@@ -419,6 +421,117 @@ def multitaper(
     return P
 
 
+def single_multitaper_spectra(
+    data,
+    alpha,
+    tapers,
+    n_f,
+    sampling_frequency,
+    nfft,
+    args_range,
+    parallel,
+):
+    """Calculate a multitaper spectrum for a single subject.
+
+    Parameters
+    ----------
+    data : np.ndarray or list
+        Raw time series data. Must have shape (n_samples, n_channels).
+    alpha : np.ndarray or list
+        Inferred mode mixing factors. Must have shape (n_samples, n_modes).
+    tapers : np.ndarray
+        Tapers for apply to each data segments.
+        Shape must be (n_tapers, segment_length).
+    n_f : int
+        Number of frequency bins.
+    sampling_frequency : float
+        Sampling frequency in Hz.
+    nfft : int
+        Number of data points to use in the FFT.
+    args_range : list
+        Minimum and maximum indices of the multitaper to keep.
+    parallel : bool
+        Is this function being called in parallel? Only affects whether
+        a progress bar is displayed or not.
+
+    Returns
+    -------
+    p : np.ndarray
+        Power spectra. Shape is (n_modes, n_channels, n_f).
+    c : np.ndarray
+        Coherence spectra. Shape is (n_modes, n_channels, n_channels, n_f).
+    """
+
+    # Use the mode mixing factors to get a time series for each mode
+    mode_time_series = time_series.get_mode_time_series(data, alpha)
+
+    # Number of subjects, modes, samples and channels
+    n_modes, n_samples, n_channels = mode_time_series.shape
+
+    # Number of tapers and segment length
+    n_tapers, segment_length = tapers.shape
+
+    # We will calculate the spectrum for several non-overlapping segments
+    # of the time series and return the average over these segments.
+
+    # Number of segments in the time series
+    n_segments = round(n_samples / segment_length)
+
+    # Power spectra for each mode
+    p = np.zeros([n_modes, n_channels, n_channels, n_f], dtype=np.complex64)
+    for i in range(n_modes):
+        if parallel:
+            iterator = range(n_segments)
+        else:
+            if i == 0:
+                print("Calculating spectra:")
+            iterator = trange(n_segments, desc=f"Mode {i}", ncols=98)
+
+        for j in iterator:
+            # Time series for mode i and segment j
+            time_series_segment = mode_time_series[
+                i, j * segment_length : (j + 1) * segment_length
+            ]
+
+            # If we're missing samples we pad with zeros either side of the data
+            if time_series_segment.shape[0] != segment_length:
+                n_zeros = segment_length - time_series_segment.shape[0]
+                n_padding = round(n_zeros / 2)
+                time_series_segment = np.pad(time_series_segment, n_padding)[
+                    :segment_length, n_padding:-n_padding
+                ]
+                if time_series_segment.shape[0] == segment_length - 1:
+                    time_series_segment = np.append(
+                        time_series_segment,
+                        np.zeros([1, time_series_segment.shape[1]], dtype=np.float32),
+                        axis=0,
+                    )
+
+            # Calculate the power (and cross) spectrum using the multitaper method
+            p[i] += multitaper(
+                time_series_segment,
+                sampling_frequency,
+                nfft=nfft,
+                tapers=tapers,
+                args_range=args_range,
+            )
+
+    # Normalise the power spectra
+    # NOTE: We should be normalising using sum alpha instead of sum
+    # alpha^2, but this makes a small difference, so we left it like
+    # this for consistency with the HMM-MAR toolbox.
+    sum_alpha = np.sum(alpha**2, axis=0)[..., np.newaxis, np.newaxis, np.newaxis]
+    p *= n_samples / (sum_alpha * n_tapers * n_segments)
+
+    # Coherences for each mode
+    c = coherence_spectra(p, print_message=False)
+
+    # Only need to keep the diagonal of the power spectra matrix
+    p = p[:, range(n_channels), range(n_channels)].real
+
+    return p, c
+
+
 def multitaper_spectra(
     data,
     alpha,
@@ -429,6 +542,7 @@ def multitaper_spectra(
     frequency_range=None,
     return_weights=False,
     standardize=False,
+    n_jobs=1,
 ):
     """Calculates spectra for inferred modes using a multitaper.
 
@@ -458,17 +572,23 @@ def multitaper_spectra(
         Useful for calculating the group average PSD.
     standardize : bool
         Should we standardize the data before calculating the multitaper?
+    n_jobs : int
+        Number of parallel jobs.
 
     Returns
     -------
     frequencies : np.ndarray
         Frequencies of the power spectra and coherences.
+        Shape is (n_f,).
     power_spectra : np.ndarray
         Power spectra for each mode.
+        Shape is (n_subjects, n_modes, n_channels, n_f).
     coherences : np.ndarray
         Coherences for each mode.
+        Shape is (n_subjects, n_modesl, n_channels, n_channels, n_f).
     weights : np.ndarray
         Weight for each subject-specific PSD. Only returned if return_weights=True.
+        Shape is (n_subjects,).
     """
 
     # Validation
@@ -498,8 +618,10 @@ def multitaper_spectra(
     if isinstance(data, np.ndarray):
         if alpha.shape[0] != data.shape[0]:
             raise ValueError("data and alpha must have the same shape.")
-        data = [data]
-        alpha = [alpha]
+
+        if data.ndim == 2:
+            data = [data]
+            alpha = [alpha]
 
     if segment_length is None:
         segment_length = 2 * sampling_frequency
@@ -533,77 +655,58 @@ def multitaper_spectra(
     if standardize:
         data = [(d - np.mean(d, axis=0)) / np.std(d, axis=0) for d in data]
 
-    print("Calculating power spectra")
+    # Calculate tapers so we can estimate spectra with the multitaper method
+    tapers = dpss(segment_length, NW=time_half_bandwidth, Kmax=n_tapers)
+    tapers *= np.sqrt(sampling_frequency)
+
+    if n_subjects == 1:
+        # We only have one subject so we don't need to parallelise the
+        # calculation
+        results = single_multitaper_spectra(
+            data[0],
+            alpha[0],
+            tapers,
+            n_f,
+            sampling_frequency,
+            nfft,
+            args_range,
+            parallel=False,
+        )
+        results = [results]
+
+    else:
+        # Create arguments to pass to single_multitaper_spectra, which will
+        # calculate spectra for each subject in parallel
+        args = []
+        for n in range(n_subjects):
+            args.append(
+                [
+                    data[n],
+                    alpha[n],
+                    tapers,
+                    n_f,
+                    sampling_frequency,
+                    nfft,
+                    args_range,
+                    True,
+                ]
+            )
+
+        # Calculate spectra in parallel
+        print("Calculating spectra:")
+        results = pqdm(
+            args,
+            single_multitaper_spectra,
+            n_jobs=n_jobs,
+            argument_type="args",
+            ncols=98,
+        )
+
+    # Unpack the results
     power_spectra = []
     coherences = []
-    for n in range(n_subjects):
-        if n_subjects > 1:
-            print(f"Subject {n}:")
-
-        # Use the mode mixing factors to get a time series for each mode
-        mode_time_series = time_series.get_mode_time_series(data[n], alpha[n])
-
-        # Number of subjects, modes, samples and channels
-        n_modes, n_samples, n_channels = mode_time_series.shape
-
-        # Calculate tapers so we can estimate spectra with the multitaper method
-        tapers = dpss(segment_length, NW=time_half_bandwidth, Kmax=n_tapers)
-        tapers *= np.sqrt(sampling_frequency)
-
-        # We will calculate the spectrum for several non-overlapping segments
-        # of the time series and return the average over these segments.
-
-        # Number of segments in the time series
-        n_segments = round(n_samples / segment_length)
-
-        # Power spectra for each mode
-        p = np.zeros([n_modes, n_channels, n_channels, n_f], dtype=np.complex64)
-        for i in range(n_modes):
-            for j in trange(n_segments, desc=f"Mode {i}", ncols=98):
-
-                # Time series for mode i and segment j
-                time_series_segment = mode_time_series[
-                    i, j * segment_length : (j + 1) * segment_length
-                ]
-
-                # If we're missing samples we pad with zeros either side of the data
-                if time_series_segment.shape[0] != segment_length:
-                    n_zeros = segment_length - time_series_segment.shape[0]
-                    n_padding = round(n_zeros / 2)
-                    time_series_segment = np.pad(time_series_segment, n_padding)[
-                        :segment_length, n_padding:-n_padding
-                    ]
-                    if time_series_segment.shape[0] == segment_length - 1:
-                        time_series_segment = np.append(
-                            time_series_segment,
-                            np.zeros(
-                                [1, time_series_segment.shape[1]], dtype=np.float32
-                            ),
-                            axis=0,
-                        )
-
-                # Calculate the power (and cross) spectrum using the multitaper method
-                p[i] += multitaper(
-                    time_series_segment,
-                    sampling_frequency,
-                    nfft=nfft,
-                    tapers=tapers,
-                    args_range=args_range,
-                )
-
-        # Normalise the power spectra
-        # NOTE: We should be normalising using sum alpha instead of sum
-        # alpha^2, but this makes a small difference, so we left it like
-        # this for consistency with the HMM-MAR toolbox.
-        sum_alpha = np.sum(alpha[n] ** 2, axis=0)[
-            ..., np.newaxis, np.newaxis, np.newaxis
-        ]
-        p *= n_samples / (sum_alpha * n_tapers * n_segments)
-
-        # Coherences for each mode
-        c = coherence_spectra(p, print_message=False)
-
-        # Add spectra to lists
+    for result in results:
+        p, c = result
         power_spectra.append(p)
         coherences.append(c)
 
@@ -617,6 +720,78 @@ def multitaper_spectra(
         return frequencies, np.squeeze(power_spectra), np.squeeze(coherences)
 
 
+def single_regression_spectra(
+    data,
+    alpha,
+    window_length,
+    sampling_frequency,
+    frequency_range,
+    calc_cpsd,
+    step_size,
+    n_sub_windows,
+    parallel,
+):
+    """Calculate cross-spectra for a single subject using the regression method.
+
+    Parameters
+    ----------
+    data : np.ndarray
+        Data to calculate the spectrogram for. Shape must be (n_samples, n_channels).
+    alpha : np.ndarray
+        Inferred mode mixing factors. Shape must be (n_samples, n_modes).
+    window_length : int
+        Number of data points to use when calculating the periodogram.
+    sampling_frequency : float
+        Sampling frequency in Hz.
+    calc_cpsd : bool
+        Should we calculate cross spectra?
+    step_size : int
+        Step size for shifting the window.
+    n_sub_windows : int
+        Should we split the window into a set of sub-windows and average the
+        spectra for each sub-window.
+    parallel : bool
+        Is this function being called in parallel? Only affects whether
+        a progress bar is displayed or not.
+
+    Returns
+    -------
+    t : np.ndarray
+        Time axis.
+    f : np.ndarray
+        Frequency axis.
+    coefs : np.ndarray
+        Regression coefficients.
+    intercept : np.ndarray
+        Regression intercept.
+    """
+    print_message = not parallel
+    t, f, p = spectrogram(
+        data,
+        window_length,
+        sampling_frequency,
+        frequency_range,
+        calc_cpsd=calc_cpsd,
+        step_size=step_size,
+        n_sub_windows=n_sub_windows,
+        print_progress_bar=print_message,
+    )
+    a = time_series.window_mean(
+        alpha,
+        window_length,
+        step_size=step_size,
+        n_sub_windows=n_sub_windows,
+    )
+    coefs, intercept = regression.linear(
+        a,
+        p,
+        fit_intercept=True,
+        normalize=True,
+        print_message=print_message,
+    )
+    return t, f, coefs, intercept
+
+
 def regression_spectra(
     data,
     alpha,
@@ -630,15 +805,18 @@ def regression_spectra(
     return_weights=False,
     return_coef_int=False,
     standardize=False,
+    n_jobs=1,
 ):
     """Calculates the PSD of each mode by regressing a time-varying PSD with alpha.
 
     Parameters
     ----------
     data : np.ndarray or list
-        Data to calculate a time-varying PSD for. Shape must be (n_samples, n_channels).
+        Data to calculate a time-varying PSD for. Shape must be (n_subjects,
+        n_samples, n_channels) or (n_samples, n_channels).
     alpha : np.ndarray
-        Inferred mode mixing factors. Shape must be (n_samples, n_modes).
+        Inferred mode mixing factors. Shape must be (n_subjects, n_samples,
+        n_modes) or (n_samples, n_modes).
     window_length : int
         Number samples to use in the window to calculate a PSD.
     sampling_frequency : float
@@ -663,6 +841,8 @@ def regression_spectra(
         separately for the PSDs?
     standardize : bool
         Should we standardize the data before calculating the spectrogram?
+    n_jobs : int
+        Number of parallel jobs.
 
     Returns
     -------
@@ -690,8 +870,9 @@ def regression_spectra(
             raise ValueError(
                 "data and alpha must both be lists or both be numpy arrays."
             )
-        data = [data]
-        alpha = [alpha]
+        if data.ndim == 2:
+            data = [data]
+            alpha = [alpha]
 
     if frequency_range is None:
         frequency_range = [0, sampling_frequency / 2]
@@ -716,27 +897,55 @@ def regression_spectra(
     # Remove the data points lost due to separating into sequences
     data = [d[: a.shape[0]] for d, a in zip(data, alpha)]
 
-    # Calculate a time-varying PSD and regress to get the mode PSDs
-    Pj = []
-    for i in range(n_subjects):
-        if n_subjects > 1:
-            print(f"Subject {i}:")
-        t, f, p = spectrogram(
-            data[i],
+    if n_subjects == 1:
+        # We only have one subject so we don't need to parallelise the
+        # calculation
+        results = single_regression_spectra(
+            data[0],
+            alpha[0],
             window_length,
             sampling_frequency,
             frequency_range,
-            calc_cpsd=calc_cpsd,
-            step_size=step_size,
-            n_sub_windows=n_sub_windows,
+            calc_cpsd,
+            step_size,
+            n_sub_windows,
+            False,
         )
-        a = time_series.window_mean(
-            alpha[i],
-            window_length,
-            step_size=step_size,
-            n_sub_windows=n_sub_windows,
+        results = [results]
+
+    else:
+        # Create arguments to pass to single_regression_spectra, which will
+        # calculate spectra for each subject in parallel
+        args = []
+        for n in range(n_subjects):
+            args.append(
+                [
+                    data[n],
+                    alpha[n],
+                    window_length,
+                    sampling_frequency,
+                    frequency_range,
+                    calc_cpsd,
+                    step_size,
+                    n_sub_windows,
+                    True,
+                ]
+            )
+
+        # Calculate a time-varying PSD and regress to get the mode PSDs
+        print("Calculating power spectra")
+        results = pqdm(
+            args,
+            single_regression_spectra,
+            n_jobs=n_jobs,
+            argument_type="args",
+            ncols=98,
         )
-        coefs, intercept = regression.linear(a, p, fit_intercept=True, normalize=True)
+
+    # Unpack results
+    Pj = []
+    for result in results:
+        t, f, coefs, intercept = result
         Pj.append([coefs, [intercept] * coefs.shape[0]])
     Pj = np.array(Pj)
 
@@ -801,6 +1010,7 @@ def spectrogram(
     calc_cpsd=True,
     step_size=1,
     n_sub_windows=1,
+    print_progress_bar=True,
 ):
     """Calculates a spectogram.
 
@@ -822,6 +1032,8 @@ def spectrogram(
     n_sub_windows : int
         Should we split the window into a set of sub-windows and average the
         spectra for each sub-window.
+    print_progress_bar : bool
+        Should we print a progress bar?
 
     Returns
     -------
@@ -875,8 +1087,11 @@ def spectrogram(
         XY_sub_window = np.empty(
             [n_sub_windows, n_channels * (n_channels + 1) // 2, n_f], dtype=np.complex64
         )
-        for i in trange(n_psds, desc="Calculating spectrogram", ncols=98):
-
+        if print_progress_bar:
+            iterator = trange(n_psds, desc="Calculating spectrogram", ncols=98)
+        else:
+            iterator = range(n_psds)
+        for i in iterator:
             # Data in the window
             j = time_indices[i]
             x_window = data[j : j + window_length].T
@@ -907,8 +1122,11 @@ def spectrogram(
         # Calculate the periodogram for each segment of the data
         P = np.empty([n_psds, n_channels, n_f], dtype=np.float32)
         XX_sub_window = np.empty([n_sub_windows, n_channels, n_f], dtype=np.float32)
-        for i in trange(n_psds, desc="Calculating spectrogram", ncols=98):
-
+        if print_progress_bar:
+            iterator = trange(n_psds, desc="Calculating spectrogram", ncols=98)
+        else:
+            iterator = range(n_psds)
+        for i in iterator:
             # Data in the window
             j = time_indices[i]
             x_window = data[j : j + window_length].T
