@@ -12,6 +12,33 @@ import osl_dynamics.inference.initializers as osld_initializers
 tfb = tfp.bijectors
 
 
+@tf.function
+def add_epsilon(A, epsilon, diag=True):
+    """Adds epsilon the the diagonal of batches of square matrices
+    or all elements of matrices.
+
+    Parameters
+    ----------
+    A : tf.Tensor
+        Batches of square matrices or vectors. Shape is (..., N, N) or (..., N)
+    epsilon : float
+        Small error added to the diagonal of the matrices or every element of the vectors.
+    diag : bool
+        Do we want to add epsilon to the diagonal?
+    """
+    # Make sure epsilon is float32
+    epsilon = tf.cast(epsilon, dtype=tf.float32)
+    A_shape = tf.shape(A)
+    if diag:
+        e = tf.eye(A_shape[-1])
+
+    else:
+        # Add epsilon to all elements
+        e = 1.0
+
+    return A + epsilon * e
+
+
 def NormalizationLayer(norm_type, *args, **kwargs):
     """Returns a normalization layer.
 
@@ -145,9 +172,14 @@ class ZeroLayer(layers.Layer):
 class InverseCholeskyLayer(layers.Layer):
     """Layer for getting Cholesky vectors from postive definite symmetric matrices."""
 
+    def __init__(self, epsilon, **kwargs):
+        super().__init__(**kwargs)
+        self.epsilon = epsilon
+        self.bijector = tfb.Chain([tfb.CholeskyOuterProduct(), tfb.FillScaleTriL()])
+
     def call(self, inputs):
-        bijector = tfb.Chain([tfb.CholeskyOuterProduct(), tfb.FillScaleTriL()])
-        return bijector.inverse(inputs)
+        inputs = add_epsilon(inputs, self.epsilon)
+        return self.bijector.inverse(inputs)
 
 
 class SampleNormalDistributionLayer(layers.Layer):
@@ -155,10 +187,20 @@ class SampleNormalDistributionLayer(layers.Layer):
 
     This layer accepts the mean and the standard deviation and
     outputs samples from a normal distribution.
+
+    Parameters
+    ----------
+    epsilon : float
+        Error to add to the standard deviations for numerical stability.
     """
 
+    def __init__(self, epsilon, **kwargs):
+        super().__init__(**kwargs)
+        self.epsilon = epsilon
+
     def call(self, inputs, training=None, **kwargs):
         mu, sigma = inputs
+        sigma = add_epsilon(sigma, self.epsilon, diag=False)
         if training:
             N = tfp.distributions.Normal(loc=mu, scale=sigma)
             return N.sample()
@@ -170,7 +212,7 @@ class SampleGumbelSoftmaxDistributionLayer(layers.Layer):
     """Layer for sampling from a Gumbel-Softmax distribution."""
 
     def call(self, inputs, **kwargs):
-        gs = tfp.distributions.RelaxedOneHotCategorical(temperature=0.05, logits=inputs)
+        gs = tfp.distributions.RelaxedOneHotCategorical(temperature=0.01, logits=inputs)
         return gs.sample()
 
 
@@ -358,6 +400,8 @@ class CovarianceMatricesLayer(layers.Layer):
         Should the matrices be learnable?
     initial_value : np.ndarray
         Initial values for the matrices.
+    epsilon : float
+        Error added to the diagonal of covariances matrices for numerical stability.
     regularizer : tf.keras.regularizers.Regularizer
         Regularizer for matrices.
     """
@@ -368,6 +412,7 @@ class CovarianceMatricesLayer(layers.Layer):
         m,
         learn,
         initial_value,
+        epsilon,
         regularizer=None,
         **kwargs,
     ):
@@ -375,6 +420,7 @@ class CovarianceMatricesLayer(layers.Layer):
         self.n = n
         self.m = m
         self.learn = learn
+        self.epsilon = epsilon
 
         # Bijector used to transform learnable vectors to covariance matrices
         self.bijector = tfb.Chain([tfb.CholeskyOuterProduct(), tfb.FillScaleTriL()])
@@ -406,7 +452,7 @@ class CovarianceMatricesLayer(layers.Layer):
         elif learn:
             # Use the identity matrix with a random error
             self.flattened_cholesky_factors_initializer = (
-                osld_initializers.NormalIdentityCholeskyInitializer(mean=0, std=0.1)
+                osld_initializers.NormalIdentityCholeskyInitializer(0.1)
             )
 
         else:
@@ -418,9 +464,6 @@ class CovarianceMatricesLayer(layers.Layer):
         # Regulariser
         self.regularizer = regularizer
 
-        # Small error to add to the diagonal to help with numerical stability
-        self.eps = tf.expand_dims(1e-5 * tf.eye(m), axis=0)
-
     def build(self, input_shape):
         self.flattened_cholesky_factors = self.add_weight(
             "flattened_cholesky_factors",
@@ -433,6 +476,7 @@ class CovarianceMatricesLayer(layers.Layer):
 
     def call(self, inputs, **kwargs):
         covariances = self.bijector(self.flattened_cholesky_factors)
+        covariances = add_epsilon(covariances, self.epsilon)
         if self.regularizer is not None:
             reg = self.regularizer(covariances)
 
@@ -440,11 +484,12 @@ class CovarianceMatricesLayer(layers.Layer):
             batch_size = tf.cast(tf.shape(inputs)[0], tf.float32)
             n_batches = self.regularizer.n_batches
             scaling_factor = batch_size * n_batches
+
             reg = reg / scaling_factor
 
             self.add_loss(reg)
             self.add_metric(reg, name=self.name)
-        return covariances + self.eps
+        return covariances
 
 
 class CorrelationMatricesLayer(layers.Layer):
@@ -463,6 +508,8 @@ class CorrelationMatricesLayer(layers.Layer):
         Should the matrices be learnable?
     initial_value : np.ndarray
         Initial values for the matrices.
+    epsilon : float
+        Error added to the diagonal of correlation matrices for numerical stability.
     regularizer : tf.keras.regularizers.Regularizer
         Regularizer for matrices.
     """
@@ -473,6 +520,7 @@ class CorrelationMatricesLayer(layers.Layer):
         m,
         learn,
         initial_value,
+        epsilon,
         regularizer=None,
         **kwargs,
     ):
@@ -480,6 +528,7 @@ class CorrelationMatricesLayer(layers.Layer):
         self.n = n
         self.m = m
         self.learn = learn
+        self.epsilon = epsilon
 
         # Bijector used to transform learnable vectors to correlation matrices
         self.bijector = tfb.Chain(
@@ -513,7 +562,7 @@ class CorrelationMatricesLayer(layers.Layer):
         elif learn:
             # Use a correlation matrix with an error added
             self.flattened_cholesky_factors_initializer = (
-                osld_initializers.NormalCorrelationCholeskyInitializer(mean=0, std=0.1)
+                osld_initializers.NormalCorrelationCholeskyInitializer(0.1)
             )
 
         else:
@@ -537,6 +586,7 @@ class CorrelationMatricesLayer(layers.Layer):
 
     def call(self, inputs, **kwargs):
         correlations = self.bijector(self.flattened_cholesky_factors)
+        correlations = add_epsilon(correlations, self.epsilon)
         if self.regularizer is not None:
             reg = self.regularizer(correlations)
 
@@ -566,6 +616,8 @@ class DiagonalMatricesLayer(layers.Layer):
         Should the matrices be learnable?
     initial_value : np.ndarray
         Initial values for the matrices.
+    epsilon : float
+        Error added to the diagonal matrices for numerical stability.
     regularizer : tf.keras.regularizers.Regularizer
         Regularizer for the diagonal entries.
     """
@@ -576,6 +628,7 @@ class DiagonalMatricesLayer(layers.Layer):
         m,
         learn,
         initial_value,
+        epsilon,
         regularizer=None,
         **kwargs,
     ):
@@ -583,6 +636,7 @@ class DiagonalMatricesLayer(layers.Layer):
         self.n = n
         self.m = m
         self.learn = learn
+        self.epsilon = epsilon
 
         # Softplus transformation to ensure diagonal is positive
         self.bijector = tfb.Softplus()
@@ -612,7 +666,7 @@ class DiagonalMatricesLayer(layers.Layer):
         elif learn:
             # Use random numbers
             self.diagonals_initializer = osld_initializers.NormalDiagonalInitializer(
-                mean=0, std=0.05
+                0.05
             )
 
         else:
@@ -638,6 +692,7 @@ class DiagonalMatricesLayer(layers.Layer):
 
     def call(self, inputs, **kwargs):
         D = self.bijector(self.diagonals)
+        D = add_epsilon(D, self.epsilon, diag=False)
         if self.regularizer is not None:
             reg = self.regularizer(D)
 
@@ -734,11 +789,7 @@ class MixMatricesLayer(layers.Layer):
         D = tf.expand_dims(tf.expand_dims(D, axis=0), axis=0)
         C = tf.reduce_sum(tf.multiply(alpha, D), axis=2)
 
-        # Small error to add to the diagonal
-        eps = 1e-5 * tf.eye(tf.shape(D)[-1])
-        eps = tf.expand_dims(tf.expand_dims(eps, axis=0), axis=0)
-
-        return C + eps
+        return C
 
 
 class ConcatVectorsMatricesLayer(layers.Layer):
@@ -756,11 +807,22 @@ class LogLikelihoodLossLayer(layers.Layer):
 
     The negative log-likelihood is calculated assuming a multivariate normal
     probability density and its value is added to the loss function.
+
+    Parameters
+    ----------
+    epsilon : float
+        Error added to the covariance matrices for numerical stability.
     """
 
+    def __init__(self, epsilon, **kwargs):
+        super().__init__(**kwargs)
+        self.epsilon = epsilon
+
     def call(self, inputs):
         x, mu, sigma = inputs
 
+        sigma = add_epsilon(sigma, self.epsilon)
+
         # Multivariate normal distribution
         mvn = tfp.distributions.MultivariateNormalTriL(
             loc=mu,
@@ -826,17 +888,23 @@ class KLDivergenceLayer(layers.Layer):
 
     Parameters
     ----------
+    epsilon : float
+        Error added to the standard deviations for numerical stability.
     clip_start : int
         Index to clip the sequences inputted to this layer.
     """
 
-    def __init__(self, clip_start=0, **kwargs):
+    def __init__(self, epsilon, clip_start=0, **kwargs):
         super().__init__(**kwargs)
         self.clip_start = clip_start
+        self.epsilon = epsilon
 
     def call(self, inputs, **kwargs):
         inference_mu, inference_sigma, model_mu, model_sigma = inputs
 
+        inference_sigma = add_epsilon(inference_sigma, self.epsilon, diag=False)
+        model_sigma = add_epsilon(model_sigma, self.epsilon, diag=False)
+
         # The model network predicts one time step into the future compared to
         # the inference network. We clip the sequences to ensure we are comparing
         # the correct time points.
@@ -1053,15 +1121,19 @@ class CategoricalLogLikelihoodLossLayer(layers.Layer):
     Parameters
     ----------
     n_states : int
-        Number of states
+        Number of states.
+    epsilon : float
+        Error added to the covariances for numerical stability.
     """
 
-    def __init__(self, n_states, **kwargs):
+    def __init__(self, n_states, epsilon, **kwargs):
         super().__init__(**kwargs)
         self.n_states = n_states
+        self.epsilon = epsilon
 
     def call(self, inputs, **kwargs):
         x, mu, sigma, probs = inputs
+        sigma = add_epsilon(sigma, self.epsilon)
 
         # Log-likelihood for each state
         ll_loss = tf.zeros(shape=tf.shape(x)[:-1])
@@ -1144,11 +1216,14 @@ class SubjectMapLayer(layers.Layer):
     ----------
     which_map : str
         Which spatial map are we using? Must be one of 'means' and 'covariances'.
+    epsilon : float
+        Error added to the diagonal of covariances for numerical stability.
     """
 
-    def __init__(self, which_map, **kwargs):
+    def __init__(self, which_map, epsilon, **kwargs):
         super().__init__(**kwargs)
         self.which_map = which_map
+        self.epsilon = epsilon
         if which_map == "covariances":
             self.bijector = tfb.Chain([tfb.CholeskyOuterProduct(), tfb.FillScaleTriL()])
         elif which_map != "means":
@@ -1166,6 +1241,7 @@ class SubjectMapLayer(layers.Layer):
 
         if self.which_map == "covariances":
             subject_map = self.bijector(subject_map)
+            subject_map = add_epsilon(subject_map, self.epsilon)
 
         return subject_map
 
@@ -1205,14 +1281,22 @@ class MixSubjectEmbeddingParametersLayer(layers.Layer):
 class SubjectMapKLDivergenceLayer(layers.Layer):
     """Layer to calculate KL divergence between posterior and prior of
     subject specific deviation.
+
+    Parameters
+    ----------
+    epsilon : float
+        Error added to the standard deviations for numerical stability.
     """
 
-    def __init__(self, n_batches=1, **kwargs):
+    def __init__(self, epsilon, n_batches=1, **kwargs):
         super().__init__(**kwargs)
+        self.epsilon = epsilon
         self.n_batches = n_batches
 
     def call(self, inputs, **kwargs):
         data, inference_mu, inference_sigma, model_sigma = inputs
+        inference_sigma = add_epsilon(inference_sigma, self.epsilon, diag=False)
+        model_sigma = add_epsilon(model_sigma, self.epsilon, diag=False)
 
         prior = tfp.distributions.Normal(loc=0.0, scale=model_sigma)
         posterior = tfp.distributions.Normal(loc=inference_mu, scale=inference_sigma)
@@ -1227,3 +1311,43 @@ class SubjectMapKLDivergenceLayer(layers.Layer):
         kl_loss = tf.reduce_sum(kl_loss)
 
         return kl_loss
+
+
+class MLPLayer(layers.Layer):
+    """Multi-Layer Perceptron layer.
+
+    Parameters
+    ----------
+    n_layers : int
+        Number of layers.
+    n_units : int
+        Number of units/neurons.
+    norm_type : str
+        Normalization layer type. Can be None, "layer" or "batch".
+    act_type : str
+        Activation type.
+    drop_rate : float
+        Dropout rate.
+    """
+
+    def __init__(
+        self,
+        n_layers,
+        n_units,
+        norm_type,
+        act_type,
+        drop_rate,
+        **kwargs,
+    ):
+        super().__init__(**kwargs)
+        self.layers = []
+        for n in range(n_layers):
+            self.layers.append(layers.Dense(n_units))
+            self.layers.append(NormalizationLayer(norm_type))
+            self.layers.append(layers.Activation(act_type))
+            self.layers.append(layers.Dropout(drop_rate))
+
+    def call(self, inputs, **kwargs):
+        for layer in self.layers:
+            inputs = layer(inputs, **kwargs)
+        return inputs
