@@ -5,6 +5,7 @@
 import numpy as np
 from pathlib import Path
 from scipy.optimize import linear_sum_assignment
+
 from osl_dynamics import array_ops, analysis
 from osl_dynamics.inference import metrics
 from osl_dynamics.utils.misc import override_dict_defaults
@@ -363,6 +364,43 @@ def fractional_occupancies(mode_time_course):
     return fo
 
 
+def calc_trans_prob_matrix(state_time_course, n_states=None):
+    """Calculate subject-specific transition probability matrices.
+
+    Parameters
+    ----------
+    state_time_course : list of np.ndarray or np.ndarray
+        State time courses.
+        Shape must be (n_subjects, n_samples, n_states) or (n_samples, n_states).
+    n_states : int
+        Number of states.
+
+    Returns
+    -------
+    trans_prob : np.ndarray
+        Subject-specific transition probability matrices.
+        Shape is (n_subjects, n_states, n_states).
+    """
+    if isinstance(state_time_course, np.ndarray):
+        state_time_course = [state_time_course]
+    trans_prob = []
+    for stc in state_time_course:
+        stc = stc.argmax(axis=1)
+        vals, counts = np.unique(
+            stc[np.arange(2)[None, :] + np.arange(len(stc) - 1)[:, None]],
+            axis=0,
+            return_counts=True,
+        )
+        if n_states is None:
+            n_states = stc.max() + 1
+        tp = np.zeros((n_states, n_states))
+        tp[vals[:, 0], vals[:, 1]] = counts
+        with np.errstate(divide="ignore", invalid="ignore"):
+            tp /= tp.sum(axis=1)[:, None]
+        trans_prob.append(np.nan_to_num(tp))
+    return np.squeeze(trans_prob)
+
+
 def fano_factor(
     time_courses,
     window_lengths,
@@ -419,37 +457,43 @@ def fano_factor(
     return np.squeeze(F)
 
 
-def fit_gmm(
+def gmm_time_courses(
     time_course,
     logit_transform=True,
     standardize=True,
-    gmm_filename=None,
+    p_value=None,
+    filename=None,
     sklearn_kwargs=None,
 ):
-    """Fit a two component GMM on the mode time courses.
+    """Fit a two component GMM on the mode time courses to get a binary
+    time course.
 
     Parameters
     ----------
-    time_course : List of np.ndarray or np.ndarray
+    time_course : list of np.ndarray or np.ndarray
         Mode time courses.
     logit_transform : bool
         Should we logit transform the mode time course?
     standardize : bool
         Should we standardize the mode time course?
-    gmm_filename : str
+    p_value : float
+        Used to determine a threshold. We ensure the data points assigned
+        to the 'on' component have a probability of less than p_value of
+        belonging to the 'off' component.
+    filename : str
         Path to directory to plot the GMM fit plots.
     sklearn_kwargs : dict
         keyword arguments for sklearn's GaussianMixture.
 
     Returns
     -------
-    gmm_time_course : List of np.ndarray or np.ndarray
+    gmm_time_course : list of np.ndarray or np.ndarray
         GMM fitted mode time courses with binary entries.
     """
 
     if isinstance(time_course, list):
         # Extract positions of discontinuities
-        discontinuities = [mtc.shape[0] for mtc in time_course]
+        discontinuities = [tc.shape[0] for tc in time_course]
         time_course = np.concatenate(time_course)
     else:
         discontinuities = None
@@ -457,16 +501,16 @@ def fit_gmm(
     n_modes = time_course.shape[1]
 
     # Initialise an array to hold the gmm thresholded time course
-    gmm_time_course = np.empty(time_course.shape)
+    gmm_time_course = np.empty(time_course.shape, dtype=int)
 
     # Loop over modes
     for mode in range(n_modes):
         a = time_course[:, mode]
 
         # GMM plot filename
-        if gmm_filename is not None:
+        if filename is not None:
             plot_filename = "{fn.parent}/{fn.stem}{mode:0{w}d}{fn.suffix}".format(
-                fn=Path(gmm_filename),
+                fn=Path(filename),
                 mode=mode,
                 w=len(str(n_modes)),
             )
@@ -474,19 +518,19 @@ def fit_gmm(
             plot_filename = None
 
         # Fit the GMM
-        default_sklearn_kwargs = {"max_iter": 5000, "n_init": 5}
+        default_sklearn_kwargs = {"max_iter": 5000, "n_init": 3}
         sklearn_kwargs = override_dict_defaults(default_sklearn_kwargs, sklearn_kwargs)
-        mixture_label = analysis.gmm.fit_gaussian_mixture(
+        threshold = analysis.gmm.fit_gaussian_mixture(
             a,
-            bayesian=False,
             logit_transform=logit_transform,
             standardize=standardize,
+            p_value=p_value,
             sklearn_kwargs=sklearn_kwargs,
             plot_filename=plot_filename,
             print_message=False,
-            return_labels=True,
         )
-        gmm_time_course[:, mode] = np.array(mixture_label)
+        print(f"GMM theshold {mode}: {threshold}")
+        gmm_time_course[:, mode] = a > threshold
 
     if discontinuities is None:
         return gmm_time_course
