@@ -1,22 +1,23 @@
 """Hidden Markov Model (HMM).
 
-NOTE:
-
-- This model is still under development.
-- This model requires a C library which has been compiled for BMRC, but will need to be recompiled if running on a different computer. The C library can be found on BMRC here: /well/woolrich/projects/software/hmm_inference_libc
-
 """
 
+import sys
+import warnings
 import os.path as op
 from dataclasses import dataclass
+from pathlib import Path
 
+import numba
 import numpy as np
 import tensorflow as tf
 import tensorflow_probability as tfp
-from tensorflow.keras import layers, utils
+from tensorflow.keras import backend, layers, utils
+from numba.core.errors import NumbaWarning
 
 import osl_dynamics.data.tf as dtf
 from osl_dynamics.simulation import HMM
+from osl_dynamics.inference import initializers
 from osl_dynamics.models import dynemo_obs
 from osl_dynamics.models.mod_base import BaseModelConfig, ModelBase
 from osl_dynamics.inference.layers import (
@@ -25,15 +26,9 @@ from osl_dynamics.inference.layers import (
     CategoricalLogLikelihoodLossLayer,
 )
 
-from ctypes import c_void_p, c_double, c_int, CDLL
-from numpy.ctypeslib import ndpointer
+warnings.filterwarnings("ignore", category=NumbaWarning)
 
-# Load a C library for hidden state inference.
-# This library has has already been compiled for BMRC.
-# Please re-compile if running on a different computer with:
-# `python setup.py build`, and update the path.
-libfile = "/well/woolrich/projects/software/hmm_inference_libc/build/lib.linux-x86_64-cpython-310/hidden_state_inference.so"
-hidden_state_inference = CDLL(libfile)
+EPS = sys.float_info.epsilon
 
 
 @dataclass
@@ -58,6 +53,8 @@ class Config(BaseModelConfig):
         Initialisation for mean vectors.
     initial_covariances : np.ndarray
         Initialisation for mode covariances.
+    covariances_epsilon : float
+        Error added to mode covariances for numerical stability.
     initial_trans_prob : np.ndarray
         Initialisation for trans prob matrix
     learn_trans_prob : bool
@@ -85,6 +82,7 @@ class Config(BaseModelConfig):
     learn_covariances: bool = None
     initial_means: np.ndarray = None
     initial_covariances: np.ndarray = None
+    covariances_epsilon: float = None
 
     initial_trans_prob: np.ndarray = None
     learn_trans_prob: bool = True
@@ -102,6 +100,12 @@ class Config(BaseModelConfig):
         if self.learn_means is None or self.learn_covariances is None:
             raise ValueError("learn_means and learn_covariances must be passed.")
 
+        if self.covariances_epsilon is None:
+            if self.learn_covariances:
+                self.covariances_epsilon = 1e-6
+            else:
+                self.covariances_epsilon = 0.0
+
 
 class Model(ModelBase):
     """HMM class.
@@ -118,22 +122,10 @@ class Model(ModelBase):
         self.model = _model_structure(self.config)
 
         self.rho = 1
-        initial_trans_prob = self.config.initial_trans_prob
-        if initial_trans_prob is None:
-            initial_trans_prob = (
-                np.ones((self.config.n_states, self.config.n_states))
-                * 0.1
-                / self.config.n_states
-            )
-            np.fill_diagonal(initial_trans_prob, 0.9)
-        self.trans_prob = initial_trans_prob
+        self.set_trans_prob(self.config.initial_trans_prob)
+        self.set_state_probs_t0(self.config.state_probs_t0)
 
-        if self.config.state_probs_t0 is None:
-            self.state_probs_t0 = (
-                np.ones((self.config.n_states,)) / self.config.n_states
-            )  # state probs at time 0
-
-    def fit(self, dataset, epochs=None):
+    def fit(self, dataset, epochs=None, lr_decay=0.1, take=1, **kwargs):
         """Fit model to a dataset.
 
         Iterates between:
@@ -146,28 +138,60 @@ class Model(ModelBase):
             Training dataset.
         epochs : int
             Number of epochs.
+        lr_decay : float
+            We update the learning rate (lr) for the observation model as
+            lr = config.learning_rate * exp(-lr_decay * epoch).
+        take : float
+            Fraction of total batches to take.
+        kwargs : keyword arguments
+            Keyword arguments for the TensorFlow observation model training.
+            These keywords arguments will be passed to self.model.fit().
 
         Returns
         -------
         history : dict
-            Dictionary with loss and rho history. Keys are 'loss' and 'rho'.
+            Dictionary with history of the loss and learning rates (lr and rho).
         """
         if epochs is None:
             epochs = self.config.n_epochs
 
+        # Make a TensorFlow Dataset
         dataset = self.make_dataset(dataset, shuffle=True, concatenate=True)
 
-        history = {"loss": [], "rho": []}
+        if take != 1:
+            # Print a message stating how many batches we'll use
+            n_total_batches = dtf.get_n_batches(dataset)
+            n_batches = max(round(n_total_batches * take), 1)
+            print(f"Using {n_batches} out of {n_total_batches} batches")
+
+        history = {"loss": [], "rho": [], "lr": []}
         for n in range(epochs):
+            if n == epochs - 1:
+                # If it's the last epoch, we train on the full dataset
+                take = 1
+
+            # Get the training data for this epoch
+            if take != 1:
+                dataset.shuffle(100000)
+                n_batches = max(round(n_total_batches * take), 1)
+                data_subset = dataset.take(n_batches)
+            else:
+                data_subset = dataset
+
+            # Setup a progress bar
             print("Epoch {}/{}".format(n + 1, epochs))
-            pb_i = utils.Progbar(len(dataset))
+            pb_i = utils.Progbar(len(data_subset))
 
             # Update rho
             self._update_rho(n)
 
+            # Set learning rate for the observation model
+            lr = self.config.learning_rate * np.exp(-lr_decay * n)
+            backend.set_value(self.model.optimizer.lr, lr)
+
             # Loop over batches
             loss = []
-            for data in dataset:
+            for data in data_subset:
                 x = data["data"]
 
                 # Update state probabilities
@@ -177,16 +201,24 @@ class Model(ModelBase):
                 if self.config.learn_trans_prob:
                     self._update_trans_prob(gamma, xi)
 
+                # Reshape gamma: (batch_size*sequence_length, n_states)
+                # -> (batch_size, sequence_length, n_states)
+                gamma = gamma.reshape(x.shape[0], x.shape[1], -1)
+
                 # Update observation model
                 training_data = np.concatenate([x, gamma], axis=2)
-                h = self.model.fit(training_data, epochs=1, verbose=0)
+                h = self.model.fit(training_data, epochs=1, verbose=0, **kwargs)
 
                 l = h.history["loss"][0]
+                if np.isnan(l):
+                    print("\nTraining failed!")
+                    return
                 loss.append(l)
-                pb_i.add(1, values=[("loss", l)])
+                pb_i.add(1, values=[("rho", self.rho), ("lr", lr), ("loss", l)])
 
             history["loss"].append(np.mean(loss))
             history["rho"].append(self.rho)
+            history["lr"].append(lr)
 
         return history
 
@@ -241,16 +273,97 @@ class Model(ModelBase):
             training_data.shuffle(100000)
             training_data_subset = training_data.take(n_batches)
             history = self.fit(training_data_subset, epochs=n_epochs, **kwargs)
+            if history is None:
+                continue
             loss = history["loss"][-1]
             if loss < best_loss:
                 best_initialization = n
                 best_loss = loss
                 best_history = history
                 best_weights = self.get_weights()
+                best_trans_prob = self.trans_prob
+
+        if best_loss == np.Inf:
+            print("Initialization failed")
+            return
 
         print(f"Using initialization {best_initialization}")
         self.reset()
-        self.set_weights(best_weights)
+        self.set_weights(best_weights, best_trans_prob)
+
+        return best_history
+
+    def random_state_time_course_initialization(
+        self, training_data, n_epochs, n_init, take=1, **kwargs
+    ):
+        """Random state time course initialization.
+
+        The model is trained for a few epochs with a sampled state time course
+        initialization. The model with the best free energy is kept.
+
+        Parameters
+        ----------
+        training_data : tensorflow.data.Dataset or osl_dynamics.data.Data
+            Dataset to use for training.
+        n_epochs : int
+            Number of epochs to train the model.
+        n_init : int
+            Number of initializations.
+        take : float
+            Fraction of total batches to take.
+        kwargs : keyword arguments
+            Keyword arguments for the fit method.
+
+        Returns
+        -------
+        history : history
+            The training history of the best initialization.
+        """
+        if n_init is None or n_init == 0:
+            print(
+                "Number of initializations was set to zero. "
+                + "Skipping initialization."
+            )
+            return
+
+        print("Random state time course initialization:")
+
+        # Make a TensorFlow Dataset
+        training_dataset = self.make_dataset(
+            training_data, shuffle=True, concatenate=True
+        )
+
+        # Calculate the number of batches to use
+        n_total_batches = dtf.get_n_batches(training_dataset)
+        n_batches = max(round(n_total_batches * take), 1)
+        print(f"Using {n_batches} out of {n_total_batches} batches")
+
+        # Pick the initialization with the lowest free energy
+        best_loss = np.Inf
+        for n in range(n_init):
+            print(f"Initialization {n}")
+            self.reset()
+            self.set_random_state_time_course_initialization(training_data)
+            training_dataset.shuffle(100000)
+            training_data_subset = training_dataset.take(n_batches)
+            history = self.fit(training_data_subset, epochs=n_epochs, **kwargs)
+            if history is None:
+                continue
+            loss = history["loss"][-1]
+            if loss < best_loss:
+                best_initialization = n
+                best_loss = loss
+                best_history = history
+                best_weights = self.get_weights()
+                best_trans_prob = self.trans_prob
+
+        if best_loss == np.Inf:
+            print("Initialization failed")
+            return
+
+        print(f"Using initialization {best_initialization}")
+        self.reset()
+        self.set_weights(best_weights, best_trans_prob)
 
         return best_history
 
@@ -261,7 +374,6 @@ class Model(ModelBase):
         ----------
         x : np.ndarray
             Observed data. Shape is (batch_size, sequence_length, n_channels).
-
         Returns
         -------
         gamma : np.ndarray
@@ -269,73 +381,73 @@ class Model(ModelBase):
             Shape is (batch_size*sequence_length, n_states).
         xi : np.ndarray
             Probability of hidden state given child and parent states, given data.
-            Shape is (batch_size*sequence_length - 1, n_states*n_states).
+            Shape is (batch_size*sequence_length - 1, n_states, n_states).
         """
-        # Would be order>0 if observation model is MAR for example,
-        # for MVN observation model order=0
-        order = 0
 
-        likelihood = self._get_likelihood(x)
-        n_samples = likelihood.shape[1] * likelihood.shape[2]
-        likelihood = np.reshape(likelihood, [-1, n_samples])  # (n_states, n_samples)
+        # Use Baum-Welch algorithm to calculate gamma, xi
+        B = self._get_likelihood(x)
+        Pi_0 = self.state_probs_t0
+        P = self.trans_prob
 
-        # Outputs
-        gamma = np.zeros(((n_samples - order), self.config.n_states))
-        xi = np.zeros(
-            ((n_samples - 1 - order), self.config.n_states * self.config.n_states)
-        )
-        scale = np.zeros((n_samples, 1))
+        gamma, xi = self._baum_welch(B, Pi_0, P)
 
-        # We need to have numpy contiguous arrays (items in array are in contiguous
-        # locations in memory) to pass into the C++ update function
-        gamma_cont = np.ascontiguousarray(gamma.T, np.double)
-        xi_cont = np.ascontiguousarray(xi.T, np.double)
-        scale_cont = np.ascontiguousarray(scale.T, np.double)
+        return gamma, xi
 
-        # Inputs
-        # We need to have numpy contiguous arrays (items in array are in contiguous
-        # locations in memory) to pass into the C++ update function
-        B_cont = np.ascontiguousarray(likelihood, np.double)
-        Pi_0_cont = np.ascontiguousarray(self.state_probs_t0.T, np.double)
-        trans_prob_cont = np.ascontiguousarray(self.trans_prob.T, np.double)
+    @numba.jit
+    def _baum_welch(self, B, Pi_0, P):
+        """Hidden state inference using the Baum-Welch algorithm.
 
-        # Define C++ class initialiser outputs types
-        hidden_state_inference.new_inferer.restype = c_void_p
+        Parameters
+        ----------
+        B : np.ndarray
+            Probability of individual data points, under observation model for
+            each state. Shape is (n_states, n_samples).
+        Pi_0 : np.ndarray
+            Initial state probabilities. Shape is (n_states,).
+        P : np.ndarray
+            State transition probabilities. Shape is (n_states, n_states).
 
-        # Define input types
-        hidden_state_inference.new_inferer.argtypes = [c_int, c_int]
+        Returns
+        -------
+        gamma : np.ndarray
+            Probability of hidden state given data. Shape is (n_samples, n_states).
+        xi : np.ndarray
+            Probability of hidden state given child and parent states, given data.
+            Shape is (n_samples - 1, n_states*n_states).
+        """
+        n_samples = B.shape[1]
+        n_states = B.shape[0]
 
-        inferer_obj = hidden_state_inference.new_inferer(self.config.n_states, order)
+        # Memory allocation
+        alpha = np.empty((n_samples, n_states))
+        beta = np.empty((n_samples, n_states))
+        scale = np.empty(n_samples)
 
-        hidden_state_inference.state_inference.argtypes = [
-            c_void_p,
-            ndpointer(dtype=c_double, shape=B_cont.shape),
-            ndpointer(dtype=c_double, shape=Pi_0_cont.shape),
-            ndpointer(dtype=c_double, shape=trans_prob_cont.shape),
-            c_int,
-            ndpointer(dtype=c_double, shape=gamma_cont.shape),
-            ndpointer(dtype=c_double, shape=xi_cont.shape),
-            ndpointer(dtype=c_double, shape=scale_cont.shape),
-        ]
+        # Forward pass
+        alpha[0] = Pi_0 * B[:, 0]
+        scale[0] = np.sum(alpha[0])
+        alpha[0] /= scale[0] + EPS
+        for i in range(1, n_samples):
+            alpha[i] = (alpha[i - 1] @ P) * B[:, i]
+            scale[i] = np.sum(alpha[i])
+            alpha[i] /= scale[i] + EPS
 
-        # Uses Baum-Welch algorithm to update gamma_cont, xi_cont, scale_cont
-        hidden_state_inference.state_inference(
-            inferer_obj,
-            B_cont,
-            Pi_0_cont,
-            trans_prob_cont,
-            n_samples,
-            gamma_cont,
-            xi_cont,
-            scale_cont,
-        )
+        # Backward pass
+        beta[-1] = 1.0 / (scale[-1] + EPS)
+        for i in range(2, n_samples + 1):
+            beta[-i] = (beta[-i + 1] * B[:, -i + 1]) @ P.T
+            beta[-i] /= scale[-i] + EPS
 
-        # Reshape gamma: (n_states, batch_size*sequence_length)
-        # -> (batch_size, sequence_length, n_states)
-        gamma = np.reshape(gamma_cont, [-1, x.shape[0], x.shape[1]])
-        gamma = np.transpose(gamma, [1, 2, 0])
+        # Marginal probabilities
+        gamma = alpha * beta
+        gamma /= np.sum(gamma, axis=1, keepdims=True)
 
-        return gamma, xi_cont
+        b = beta[1:] * B[:, 1:].T
+        t = P * np.expand_dims(alpha[:-1], axis=2) * np.expand_dims(b, axis=1)
+        xi = t.reshape(n_samples - 1, -1, order="F")
+        xi /= np.expand_dims(np.sum(xi, axis=1), axis=1) + EPS
+
+        return gamma, xi
 
     def _get_likelihood(self, x):
         """Get likelihood time series.
@@ -347,24 +459,38 @@ class Model(ModelBase):
 
         Returns
         -------
-        np.ndarray
-            Likelihood time series. Shape is (n_states, batch_size, sequence_length).
+        likelihood : np.ndarray
+            Likelihood time series. Shape is (n_states, batch_size*sequence_length).
         """
+
+        # Get the current observation model parameters
         means, covs = self.get_means_covariances()
 
         n_states = means.shape[0]
         batch_size = x.shape[0]
         sequence_length = x.shape[1]
+        n_samples = batch_size * sequence_length
 
-        log_likelihood = np.empty([n_states, batch_size, sequence_length])
+        # Calculate the log-likelihood for each state to have generated the
+        # observed data
+        log_likelihood = np.empty([n_states, n_samples])
         for state in range(n_states):
             mvn = tfp.distributions.MultivariateNormalTriL(
                 loc=means[state],
                 scale_tril=tf.linalg.cholesky(covs[state]),
                 allow_nan_stats=False,
             )
-            log_likelihood[state] = mvn.log_prob(x)
+            log_likelihood[state] = np.reshape(mvn.log_prob(x), n_samples)
 
+        # We add a constant to the log-likelihood for time points where all states
+        # have a negative log-likelihood. This is critical for numerical stability.
+        time_points_with_all_states_negative = np.all(log_likelihood < 0, axis=0)
+        if np.any(time_points_with_all_states_negative):
+            log_likelihood[:, time_points_with_all_states_negative] -= np.max(
+                log_likelihood[:, time_points_with_all_states_negative], axis=0
+            )
+
+        # Return the likelihood
         return np.exp(log_likelihood)
 
     def _update_trans_prob(self, gamma, xi):
@@ -377,16 +503,12 @@ class Model(ModelBase):
             Shape is (batch_size*sequence_length, n_states).
         xi : np.ndarray
             Probability of hidden state given child and parent states, given data.
-            Shape is (batch_size*sequence_length - 1, n_states, n_states).
+            Shape is (batch_size*sequence_length - 1, n_states*n_states).
         """
-        # Reshape gamma: (batch_size, sequence_length, n_states)
-        # -> (n_states, batch_size*sequence_length)
-        gamma = gamma.reshape(-1, gamma.shape[-1]).T
-
         # Use Baum-Welch algorithm
-        phi_interim = np.reshape(
-            np.sum(xi, 1), [self.config.n_states, self.config.n_states]
-        ).T / np.reshape(np.sum(gamma[:, :-1], 1), [self.config.n_states, 1])
+        phi_interim = np.sum(xi, axis=0).reshape(
+            self.config.n_states, self.config.n_states
+        ).T / np.sum(gamma[:-1], axis=0).reshape(self.config.n_states, 1)
 
         # We use stochastic updates on trans_prob as per Eqs. (1) and (2) in the paper:
         # https://www.sciencedirect.com/science/article/pii/S1053811917305487
@@ -418,7 +540,7 @@ class Model(ModelBase):
 
         Returns
         -------
-        np.ndarray
+        stc : np.ndarray
             State time course with shape (n_samples, n_states).
         """
         sim = HMM(self.trans_prob)
@@ -430,7 +552,8 @@ class Model(ModelBase):
 
         Returns
         -------
-        np.ndarray
+        trans_prob : np.ndarray
+            Transition probability matrix. Shape is (n_states, n_states).
         """
         return self.trans_prob
 
@@ -439,7 +562,8 @@ class Model(ModelBase):
 
         Returns
         -------
-        np.ndarary
+        covariances : np.ndarray
+            State covariances. Shape is (n_states, n_channels, n_channels).
         """
         return dynemo_obs.get_covariances(self.model)
 
@@ -481,6 +605,79 @@ class Model(ModelBase):
         """
         dynemo_obs.set_covariances(self.model, covariances, update_initializer)
 
+    def set_trans_prob(self, trans_prob):
+        """Sets the transition probability matrix.
+
+        Parameters
+        ----------
+        trans_prob : np.ndarray
+            State transition probabilities. Shape must be (n_states, n_states).
+        """
+        if trans_prob is None:
+            trans_prob = (
+                np.ones((self.config.n_states, self.config.n_states))
+                * 0.1
+                / (self.config.n_states - 1)
+            )
+            np.fill_diagonal(trans_prob, 0.9)
+        self.trans_prob = trans_prob
+
+    def set_state_probs_t0(self, state_probs_t0):
+        """Set the initial state probabilities.
+
+        Parameters
+        ----------
+        state_probs_t0 : np.ndarray
+            Initial state probabilities. Shape is (n_states,).
+        """
+
+        if state_probs_t0 is None:
+            state_probs_t0 = np.ones((self.config.n_states,)) / self.config.n_states
+        self.state_probs_t0 = state_probs_t0
+
+    def set_random_state_time_course_initialization(self, training_data):
+        """Sets the initial means/covariances based on a random state time course.
+
+        Parameters
+        ----------
+        training_data : osl_dynamics.data.Data
+            Training data object.
+        """
+
+        # Loop over subjects
+        subject_means = []
+        subject_covariances = []
+        for data in training_data.subjects:
+
+            # Sample a state time course using the initial transition
+            # probability matrix
+            stc = self.sample_state_time_course(data.shape[0])
+
+            # Calculate the mean/covariance for each state for this subject
+            m = []
+            C = []
+            for j in range(self.config.n_states):
+                x = data[stc[:, j] == 1]
+                mu_j = np.mean(x, axis=0)
+                sigma_j = np.cov(x, rowvar=False)
+                m.append(mu_j)
+                C.append(sigma_j)
+
+            subject_means.append(m)
+            subject_covariances.append(C)
+
+        # Average over subjects
+        initial_means = np.mean(subject_means, axis=0)
+        initial_covariances = np.mean(subject_covariances, axis=0)
+
+        if self.config.learn_means:
+            # Set initial means
+            self.set_means(initial_means, update_initializer=True)
+
+        if self.config.learn_covariances:
+            # Set initial covariances
+            self.set_covariances(initial_covariances, update_initializer=True)
+
     def set_regularizers(self, training_dataset):
         """Set the means and covariances regularizer based on the training data.
 
@@ -513,18 +710,20 @@ class Model(ModelBase):
 
         Returns
         -------
-        list or np.ndarray
+        alpha : list or np.ndarray
             State probabilities with shape (n_subjects, n_samples, n_states)
             or (n_samples, n_states).
         """
         dataset = self.make_dataset(dataset, concatenate=False)
 
+        print("Getting alpha")
         alpha = []
         for ds in dataset:
             gamma = []
             for data in ds:
-                g, _ = self._get_state_probs(data["data"])
-                gamma.append(np.concatenate(g))
+                x = data["data"]
+                g, _ = self._get_state_probs(x)
+                gamma.append(g)
             alpha.append(np.concatenate(gamma))
 
         if concatenate or len(alpha) == 1:
@@ -553,27 +752,45 @@ class Model(ModelBase):
             self.config.sequence_length, prepared=prepared, concatenate=concatenate
         )
 
-    def save_weights(self, dirname):
+    def save_weights(self, filepath):
         """Save all model weights.
 
         Parameters
         ----------
-        dirname : str
-            Location to save model parameters to.
+        filepath : str
+            Location to save model weights to.
         """
-        self.model.save_weights(op.join(dirname, "weights"))
-        np.save(op.join(dirname, "trans_prob.npy"), self.trans_prob)
+        self.model.save_weights(filepath)
+        np.save(op.join(str(Path(filepath).parent), "trans_prob.npy"), self.trans_prob)
 
-    def load_weights(self, dirname):
+    def load_weights(self, filepath):
         """Load all model parameters.
 
         Parameters
         ----------
-        dirname : str
-            Location to load model parameters from.
+        filepath : str
+            Location to load model weights from.
         """
-        self.model.load_weights(op.join(dirname, "weights"))
-        self.trans_prob = np.load(op.join(dirname, "trans_prob.npy"))
+        self.trans_prob = np.load(op.join(str(Path(filepath).parent), "trans_prob.npy"))
+        return self.model.load_weights(filepath)
+
+    def set_weights(self, weights, trans_prob):
+        """Set model parameter weights.
+
+        Parameters
+        ----------
+        weights : tensorflow weights
+            TensorFlow weights for the observation model.
+        trans_prob : np.ndarray
+            Transition probability matrix.
+        """
+        self.model.set_weights(weights)
+        self.set_trans_prob(trans_prob)
+
+    def reset_weights(self):
+        """Resets trainable variables in the model to their initial value."""
+        initializers.reinitialize_model_weights(self.model)
+        self.set_trans_prob(self.config.initial_trans_prob)
 
 
 def _model_structure(config):
@@ -582,7 +799,7 @@ def _model_structure(config):
         shape=(config.sequence_length, config.n_channels + config.n_states),
         name="inputs",
     )
-    data, gamma = tf.split(inputs, [config.n_channels, config.n_states], 2)
+    data, gamma = tf.split(inputs, [config.n_channels, config.n_states], axis=2)
 
     # Definition of layers
     means_layer = MeanVectorsLayer(
@@ -597,9 +814,12 @@ def _model_structure(config):
         config.n_channels,
         config.learn_covariances,
         config.initial_covariances,
+        config.covariances_epsilon,
         name="covs",
     )
-    ll_loss_layer = CategoricalLogLikelihoodLossLayer(config.n_states, name="ll_loss")
+    ll_loss_layer = CategoricalLogLikelihoodLossLayer(
+        config.n_states, config.covariances_epsilon, name="ll_loss"
+    )
 
     # Data flow
     mu = means_layer(data)  # data not used
