@@ -98,6 +98,10 @@ class Config(BaseModelConfig):
         Initialisation for mean vectors.
     initial_covariances : np.ndarray
         Initialisation for mode covariances.
+    stds_epsilon : float
+        Error added to mode stds for numerical stability.
+    fcs_epsilon : float
+        Error added to mode fcs for numerical stability.
 
     batch_size : int
         Mini-batch size.
@@ -149,6 +153,8 @@ class Config(BaseModelConfig):
     initial_means: np.ndarray = None
     initial_stds: np.ndarray = None
     initial_fcs: np.ndarray = None
+    stds_epsilon: float = None
+    fcs_epsilon: float = None
     multiple_dynamics: bool = True
 
     def __post_init__(self):
@@ -175,6 +181,18 @@ class Config(BaseModelConfig):
         ):
             raise ValueError("learn_means, learn_stds and learn_fcs must be passed.")
 
+        if self.stds_epsilon is None:
+            if self.learn_stds:
+                self.stds_epsilon = 1e-6
+            else:
+                self.stds_epsilon = 0.0
+
+        if self.fcs_epsilon is None:
+            if self.learn_fcs:
+                self.fcs_epsilon = 1e-6
+            else:
+                self.fcs_epsilon = 0.0
+
 
 class Model(ModelBase):
     """MAGE model class.
@@ -576,6 +594,7 @@ def _build_inference_model(config):
         config.n_channels,
         config.learn_stds,
         config.initial_stds,
+        config.stds_epsilon,
         name="stds",
     )
 
@@ -584,6 +603,7 @@ def _build_inference_model(config):
         config.n_channels,
         config.learn_fcs,
         config.initial_fcs,
+        config.fcs_epsilon,
         name="fcs",
     )
 
