@@ -29,13 +29,12 @@ from osl_dynamics.inference.layers import (
     ConcatenateLayer,
     SubjectDevEmbeddingLayer,
     SubjectMapLayer,
-    DummyLayer,
     MixSubjectEmbeddingParametersLayer,
     TFRangeLayer,
     ZeroLayer,
     InverseCholeskyLayer,
-    ScalarLayer,
     SubjectMapKLDivergenceLayer,
+    MLPLayer,
 )
 
 
@@ -103,6 +102,8 @@ class Config(BaseModelConfig, VariationalInferenceModelConfig):
         Initialisation for mean vectors.
     initial_covariances : np.ndarray
         Initialisation for mode covariances.
+    covariances_epsilon : float
+        Error added to mode covariances for numerical stability.
     means_regularizer : tf.keras.regularizers.Regularizer
         Regularizer for group mean vectors.
     covariances_regularizer : tf.keras.regularizers.Regularizer
@@ -140,18 +141,29 @@ class Config(BaseModelConfig, VariationalInferenceModelConfig):
         Number of dimensions for the subject embedding.
     mode_embedding_dim : int
         Number of dimensions for the mode embedding in the spatial maps encoder.
-    mode_embedding_activation : str
-        Activation for the mode encoders.
-    dev_reg : str
-        Type of regularisation applied to the deviations.
-    dev_reg_strength : float
-        Strength of regularisation to deviations.
-    dev_bayesian : bool
-        Do we want to be fully Bayesian on deviations?
-    learn_dev_mod_sigma : bool
-        Do we want to learn the prior std of the deviation.
-    initial_dev_mod_sigma : float
-        Initial value for prior std of the deviation.
+
+    dev_inf_n_layers : int
+        Number of layers for the inference MLP for deviations.
+    dev_inf_n_units : int
+        Number of units for the inference MLP for deviations.
+    dev_inf_normalization : str
+        Type of normalization for the inference MLP for deviations.
+        Either None, 'batch' or 'layer'.
+    dev_inf_activation : str
+        Type of activation to use for the inference MLP for deviations.
+        E.g. 'relu', 'sigmoid', 'tanh', etc.
+    dev_inf_dropout : float
+        Dropout rate for the inference MLP for deviations.
+    dev_mod_n_layers : int
+        Number of layers for the model MLP for deviations.
+    dev_mod_n_units : int
+        Number of units for the model MLP for deviations.
+    dev_mod_normalization : str
+        Type of normalization for the model MLP for deviations.
+    dev_mod_activation : str
+        Type of activation to use for the model MLP for deviations.
+    dev_mod_dropout : float
+        Dropout rate for the model MLP for deviations.
     """
 
     model_name: str = "SE-DyNeMo"
@@ -179,6 +191,7 @@ class Config(BaseModelConfig, VariationalInferenceModelConfig):
     learn_covariances: bool = None
     initial_means: np.ndarray = None
     initial_covariances: np.ndarray = None
+    covariances_epsilon: float = None
     means_regularizer: tf.keras.regularizers.Regularizer = None
     covariances_regularizer: tf.keras.regularizers.Regularizer = None
 
@@ -186,12 +199,17 @@ class Config(BaseModelConfig, VariationalInferenceModelConfig):
     n_subjects: int = None
     subject_embedding_dim: int = None
     mode_embedding_dim: int = None
-    mode_embedding_activation: str = None
-    dev_reg: str = None
-    dev_reg_strength: float = 0.0
-    dev_bayesian: bool = False
-    learn_dev_mod_sigma: bool = True
-    initial_dev_mod_sigma: float = 1
+
+    dev_inf_n_layers: int = 0
+    dev_inf_n_units: int = None
+    dev_inf_normalization: str = None
+    dev_inf_activation: str = None
+    dev_inf_dropout: float = 0.0
+    dev_mod_n_layers: int = 0
+    dev_mod_n_units: int = None
+    dev_mod_normalization: str = None
+    dev_mod_activation: str = None
+    dev_mod_dropout: float = 0.0
 
     def __post_init__(self):
         self.validate_rnn_parameters()
@@ -213,6 +231,12 @@ class Config(BaseModelConfig, VariationalInferenceModelConfig):
         if self.learn_means is None or self.learn_covariances is None:
             raise ValueError("learn_means and learn_covariances must be passed.")
 
+        if self.covariances_epsilon is None:
+            if self.learn_covariances:
+                self.covariances_epsilon = 1e-6
+            else:
+                self.covariances_epsilon = 0.0
+
     def validate_subject_embedding_parameters(self):
         if (
             self.n_subjects is None
@@ -223,6 +247,12 @@ class Config(BaseModelConfig, VariationalInferenceModelConfig):
                 "n_subjects, subject_embedding_dim and mode_embedding_dim must be passed."
             )
 
+        if self.dev_inf_n_layers != 0 and self.dev_inf_n_units is None:
+            raise ValueError("Please pass dev_inf_n_units.")
+
+        if self.dev_mod_n_layers != 0 and self.dev_mod_n_units is None:
+            raise ValueError("Plase pass dev_mod_n_units.")
+
 
 class Model(VariationalInferenceModelBase):
     """Subject Embedded RNN Inference/model network and Gaussian observatons (SERIGO).
@@ -300,7 +330,7 @@ class Model(VariationalInferenceModelBase):
             Deviation of Cholesky factor of covs from the group.
             Shaoe is (n_subjects, n_modes, n_channels * (n_channels + 1) // 2).
         """
-        return sedynemo_obs.get_subject_dev(self.model, self.config.dev_bayesian)
+        return sedynemo_obs.get_subject_dev(self.model)
 
     def get_subject_means_covariances(self):
         """Get the means and covariances for each subject
@@ -312,9 +342,7 @@ class Model(VariationalInferenceModelBase):
         subject_covs : np.ndarray
             Mode covariances for each subject. Shape is (n_subjects, n_modes, n_channels, n_channels).
         """
-        return sedynemo_obs.get_subject_means_covariances(
-            self.model, self.config.dev_bayesian
-        )
+        return sedynemo_obs.get_subject_means_covariances(self.model)
 
     def set_regularizers(self, training_dataset):
         """Set the means and covariances regularizer based on the training data.
@@ -438,7 +466,7 @@ def _model_structure(config):
 
     # Layers to sample theta from q(theta) and to convert to mode mixing
     # factors alpha
-    theta_layer = SampleNormalDistributionLayer(name="theta")
+    theta_layer = SampleNormalDistributionLayer(config.theta_std_epsilon, name="theta")
     theta_norm_layer = NormalizationLayer(config.theta_normalization, name="theta_norm")
     alpha_layer = SoftmaxLayer(
         config.initial_alpha_temperature,
@@ -484,17 +512,16 @@ def _model_structure(config):
         config.n_channels,
         config.learn_covariances,
         config.initial_covariances,
+        config.covariances_epsilon,
         config.covariances_regularizer,
         name="group_covs",
     )
     means_mode_embedding_layer = layers.Dense(
         config.mode_embedding_dim,
-        config.mode_embedding_activation,
         name="means_mode_embedding",
     )
     covs_mode_embedding_layer = layers.Dense(
         config.mode_embedding_dim,
-        config.mode_embedding_activation,
         name="covs_mode_embedding",
     )
     means_concat_embedding_layer = SubjectDevEmbeddingLayer(
@@ -509,87 +536,70 @@ def _model_structure(config):
         config.n_subjects,
         name="covs_concat_embedding",
     )
-    # ----------------------------------------- #
-    # Layers specific to the non Bayesian model #
-    # ----------------------------------------- #
-    if not config.dev_bayesian:
-        if config.learn_means:
-            means_dev_layer = layers.Dense(config.n_channels, name="means_dev")
-        else:
-            means_dev_layer = ZeroLayer(
-                shape=(config.n_subjects, config.n_modes, config.n_channels),
-                name="means_dev",
-            )
 
-        if config.learn_covariances:
-            covs_dev_layer = layers.Dense(
-                config.n_channels * (config.n_channels + 1) // 2,
-                name="covs_dev",
-            )
-        else:
-            covs_dev_layer = ZeroLayer(
-                shape=(
-                    config.n_subjects,
-                    config.n_modes,
-                    config.n_channels * (config.n_channels + 1) // 2,
-                ),
-                name="covs_dev",
-            )
-
-        means_dev_reg_layer = DummyLayer(
-            config.dev_reg,
-            config.dev_reg_strength,
-            name="means_dev_reg",
+    # Inference part of the deviation
+    means_dev_inf_input_layer = MLPLayer(
+        config.dev_inf_n_layers,
+        config.dev_inf_n_units,
+        config.dev_inf_normalization,
+        config.dev_inf_activation,
+        config.dev_inf_dropout,
+        name="means_dev_inf_input",
+    )
+    means_dev_inf_mu_layer = layers.Dense(config.n_channels, name="means_dev_inf_mu")
+    means_dev_inf_sigma_layer = layers.Dense(
+        config.n_channels, activation="softplus", name="means_dev_inf_sigma"
+    )
+    if config.learn_means:
+        means_dev_layer = SampleNormalDistributionLayer(
+            config.theta_std_epsilon, name="means_dev"
         )
-        covs_dev_reg_layer = DummyLayer(
-            config.dev_reg,
-            config.dev_reg_strength,
-            name="covs_dev_reg",
-        )
-    # ------------------------------------- #
-    # Layers specific to the Bayesian model #
-    # ------------------------------------- #
     else:
-        means_dev_inf_mu_layer = layers.Dense(
-            config.n_channels, name="means_dev_inf_mu"
+        means_dev_layer = ZeroLayer(
+            shape=(config.n_subjects, config.n_modes, config.n_channels),
+            name="means_dev",
         )
-        means_dev_inf_sigma_layer = layers.Dense(
-            config.n_channels, activation="softplus", name="means_dev_inf_sigma"
-        )
-        if config.learn_means:
-            means_dev_layer = SampleNormalDistributionLayer(name="means_dev")
-        else:
-            means_dev_layer = ZeroLayer(
-                shape=(config.n_subjects, config.n_modes, config.n_channels),
-                name="means_dev",
-            )
 
-        covs_dev_inf_mu_layer = layers.Dense(
-            config.n_channels * (config.n_channels + 1) // 2, name="covs_dev_inf_mu"
+    covs_dev_inf_input_layer = MLPLayer(
+        config.dev_inf_n_layers,
+        config.dev_inf_n_units,
+        config.dev_inf_normalization,
+        config.dev_inf_activation,
+        config.dev_inf_dropout,
+        name="covs_dev_inf_input",
+    )
+    covs_dev_inf_mu_layer = layers.Dense(
+        config.n_channels * (config.n_channels + 1) // 2, name="covs_dev_inf_mu"
+    )
+    covs_dev_inf_sigma_layer = layers.Dense(
+        config.n_channels * (config.n_channels + 1) // 2,
+        activation="softplus",
+        name="covs_dev_inf_sigma",
+    )
+    if config.learn_covariances:
+        covs_dev_layer = SampleNormalDistributionLayer(
+            config.theta_std_epsilon, name="covs_dev"
         )
-        covs_dev_inf_sigma_layer = layers.Dense(
-            config.n_channels * (config.n_channels + 1) // 2,
-            activation="softplus",
-            name="covs_dev_inf_sigma",
+    else:
+        covs_dev_layer = ZeroLayer(
+            shape=(
+                config.n_subjects,
+                config.n_modes,
+                config.n_channels * (config.n_channels + 1) // 2,
+            ),
+            name="covs_dev",
         )
-        if config.learn_covariances:
-            covs_dev_layer = SampleNormalDistributionLayer(name="covs_dev")
-        else:
-            covs_dev_layer = ZeroLayer(
-                shape=(
-                    config.n_subjects,
-                    config.n_modes,
-                    config.n_channels * (config.n_channels + 1) // 2,
-                ),
-                name="covs_dev",
-            )
 
-    subject_means_layer = SubjectMapLayer("means", name="subject_means")
-    subject_covs_layer = SubjectMapLayer("covariances", name="subject_covs")
+    subject_means_layer = SubjectMapLayer(
+        "means", config.theta_std_epsilon, name="subject_means"
+    )
+    subject_covs_layer = SubjectMapLayer(
+        "covariances", config.theta_std_epsilon, name="subject_covs"
+    )
     mix_subject_means_covs_layer = MixSubjectEmbeddingParametersLayer(
         name="mix_subject_means_covs"
     )
-    ll_loss_layer = LogLikelihoodLossLayer(name="ll_loss")
+    ll_loss_layer = LogLikelihoodLossLayer(config.theta_std_epsilon, name="ll_loss")
 
     # Data flow
     subjects = subjects_layer(data)  # data not used here
@@ -600,7 +610,9 @@ def _model_structure(config):
 
     # spatial map embeddings
     means_mode_embedding = means_mode_embedding_layer(group_mu)
-    covs_mode_embedding = covs_mode_embedding_layer(InverseCholeskyLayer()(group_D))
+    covs_mode_embedding = covs_mode_embedding_layer(
+        InverseCholeskyLayer(config.covariances_epsilon)(group_D)
+    )
 
     # Now get the subject specific spatial maps
     means_concat_embedding = means_concat_embedding_layer(
@@ -610,20 +622,15 @@ def _model_structure(config):
         [subject_embeddings, covs_mode_embedding]
     )
 
-    if not config.dev_bayesian:
-        means_dev = means_dev_layer(means_concat_embedding)
-        means_dev = means_dev_reg_layer(means_dev)
+    means_dev_inf_input = means_dev_inf_input_layer(means_concat_embedding)
+    means_dev_inf_mu = means_dev_inf_mu_layer(means_dev_inf_input)
+    means_dev_inf_sigma = means_dev_inf_sigma_layer(means_dev_inf_input)
+    means_dev = means_dev_layer([means_dev_inf_mu, means_dev_inf_sigma])
 
-        covs_dev = covs_dev_layer(covs_concat_embedding)
-        covs_dev = covs_dev_reg_layer(covs_dev)
-    else:
-        means_dev_inf_mu = means_dev_inf_mu_layer(means_concat_embedding)
-        means_dev_inf_sigma = means_dev_inf_sigma_layer(means_concat_embedding)
-        means_dev = means_dev_layer([means_dev_inf_mu, means_dev_inf_sigma])
-
-        covs_dev_inf_mu = covs_dev_inf_mu_layer(covs_concat_embedding)
-        covs_dev_inf_sigma = covs_dev_inf_sigma_layer(covs_concat_embedding)
-        covs_dev = covs_dev_layer([covs_dev_inf_mu, covs_dev_inf_sigma])
+    covs_dev_inf_input = covs_dev_inf_input_layer(covs_concat_embedding)
+    covs_dev_inf_mu = covs_dev_inf_mu_layer(covs_dev_inf_input)
+    covs_dev_inf_sigma = covs_dev_inf_sigma_layer(covs_dev_inf_input)
+    covs_dev = covs_dev_layer([covs_dev_inf_mu, covs_dev_inf_sigma])
 
     mu = subject_means_layer([group_mu, means_dev])
     D = subject_covs_layer([group_D, covs_dev])
@@ -656,7 +663,7 @@ def _model_structure(config):
     mod_sigma_layer = layers.Dense(
         config.n_modes, activation="softplus", name="mod_sigma"
     )
-    kl_div_layer = KLDivergenceLayer(name="kl_div")
+    kl_div_layer = KLDivergenceLayer(config.theta_std_epsilon, name="kl_div")
     kl_loss_layer = KLLossLayer(config.do_kl_annealing, name="kl_loss")
 
     # Data flow
@@ -668,38 +675,62 @@ def _model_structure(config):
     mod_sigma = mod_sigma_layer(model_output_concat)
     kl_div = kl_div_layer([inf_mu, inf_sigma, mod_mu, mod_sigma])
 
-    if not config.dev_bayesian:
-        kl_loss = kl_loss_layer(kl_div)
+    # Prior part of deviations
+    means_dev_mod_input_layer = MLPLayer(
+        config.dev_inf_n_layers,
+        config.dev_inf_n_units,
+        config.dev_inf_normalization,
+        config.dev_inf_activation,
+        config.dev_inf_dropout,
+        name="means_dev_mod_input",
+    )
+    means_dev_mod_sigma_layer = layers.Dense(
+        config.n_channels, activation="softplus", name="means_dev_mod_sigma"
+    )
+    covs_dev_mod_input_layer = MLPLayer(
+        config.dev_inf_n_layers,
+        config.dev_inf_n_units,
+        config.dev_inf_normalization,
+        config.dev_inf_activation,
+        config.dev_inf_dropout,
+        name="covs_dev_mod_input",
+    )
+    covs_dev_mod_sigma_layer = layers.Dense(
+        config.n_channels * (config.n_channels + 1) // 2,
+        activation="softplus",
+        name="covs_dev_mod_sigma",
+    )
+
+    if config.learn_means:
+        means_dev_kl_loss_layer = SubjectMapKLDivergenceLayer(
+            config.theta_std_epsilon, name="means_dev_kl_loss"
+        )
     else:
-        # Layers for dev prior
-        dev_mod_sigma_layer = ScalarLayer(
-            config.learn_dev_mod_sigma,
-            config.initial_dev_mod_sigma,
-            name="dev_mod_sigma",
-        )
-        if config.learn_means:
-            means_dev_kl_loss_layer = SubjectMapKLDivergenceLayer(
-                name="means_dev_kl_loss"
-            )
-        else:
-            means_dev_kl_loss_layer = ZeroLayer((), name="means_dev_kl_loss")
+        means_dev_kl_loss_layer = ZeroLayer((), name="means_dev_kl_loss")
 
-        if config.learn_covariances:
-            covs_dev_kl_loss_layer = SubjectMapKLDivergenceLayer(
-                name="covs_dev_kl_loss"
-            )
-        else:
-            covs_dev_kl_loss_layer = ZeroLayer((), name="covs_dev_kl_loss")
+    if config.learn_covariances:
+        covs_dev_kl_loss_layer = SubjectMapKLDivergenceLayer(
+            config.theta_std_epsilon, name="covs_dev_kl_loss"
+        )
+    else:
+        covs_dev_kl_loss_layer = ZeroLayer((), name="covs_dev_kl_loss")
 
-        # Data flow
-        dev_mod_sigma = dev_mod_sigma_layer(data)  # Data not used
-        means_dev_kl_loss = means_dev_kl_loss_layer(
-            [data, means_dev_inf_mu, means_dev_inf_sigma, dev_mod_sigma]
-        )
-        covs_dev_kl_loss = covs_dev_kl_loss_layer(
-            [data, covs_dev_inf_mu, covs_dev_inf_sigma, dev_mod_sigma]
-        )
-        kl_loss = kl_loss_layer([kl_div, means_dev_kl_loss, covs_dev_kl_loss])
+    kl_loss_layer = KLLossLayer(do_annealing=config.do_kl_annealing, name="kl_loss")
+
+    # Data flow
+    means_dev_mod_input = means_dev_mod_input_layer(means_concat_embedding)
+    means_dev_mod_sigma = means_dev_mod_sigma_layer(means_dev_mod_input)
+
+    covs_dev_mod_input = covs_dev_mod_input_layer(covs_concat_embedding)
+    covs_dev_mod_sigma = covs_dev_mod_sigma_layer(covs_dev_mod_input)
+
+    means_dev_kl_loss = means_dev_kl_loss_layer(
+        [data, means_dev_inf_mu, means_dev_inf_sigma, means_dev_mod_sigma]
+    )
+    covs_dev_kl_loss = covs_dev_kl_loss_layer(
+        [data, covs_dev_inf_mu, covs_dev_inf_sigma, covs_dev_mod_sigma]
+    )
+    kl_loss = kl_loss_layer([kl_div, means_dev_kl_loss, covs_dev_kl_loss])
 
     return tf.keras.Model(
         inputs=[data, subj_id],
