@@ -13,6 +13,7 @@ import osl_dynamics.data.tf as dtf
 from osl_dynamics.models import dynemo_obs
 from osl_dynamics.models.mod_base import BaseModelConfig, ModelBase
 from osl_dynamics.inference.layers import (
+    add_epsilon,
     DummyLayer,
     LogLikelihoodLossLayer,
     MeanVectorsLayer,
@@ -24,9 +25,9 @@ from osl_dynamics.inference.layers import (
     ZeroLayer,
     InverseCholeskyLayer,
     SampleNormalDistributionLayer,
-    ScalarLayer,
     SubjectMapKLDivergenceLayer,
     KLLossLayer,
+    MLPLayer,
 )
 
 
@@ -52,6 +53,8 @@ class Config(BaseModelConfig):
         Initialisation for mean vectors.
     initial_covariances : np.ndarray
         Initialisation for mode covariances.
+    covariances_epsilon : float
+        Error added to mode covariances for numerical stability.
     means_regularizer : tf.keras.regularizers.Regularizer
         Regularizer for group mean vectors.
     covariances_regularizer : tf.keras.regularizers.Regularizer
@@ -63,20 +66,29 @@ class Config(BaseModelConfig):
         Number of dimensions for the subject embedding.
     mode_embedding_dim : int
         Number of dimensions for the mode embedding in the spatial maps encoder.
-    mode_embedding_activation : str
-        Activation of the mode encoders.
-    dev_reg : str
-        Type of regularisation applied to the deviations.
-    dev_reg_strength : float
-        Strength of regularisation to deviations.
-    dev_bayesian : bool
-        Do we want to be fully Bayesian on deviations?.
-    learn_dev_mod_sigma : bool
-        Do we want to learn the prior std of the deviation.
-    initial_dev_mod_sigma : float
-        Initial value for prior std of the deviation.
-    do_kl_annealing : bool
-        Should we do annealing on the KL loss during training?
+
+    dev_inf_n_layers : int
+        Number of layers for the inference MLP for deviations.
+    dev_inf_n_units : int
+        Number of units for the inference MLP for deviations.
+    dev_inf_normalization : str
+        Type of normalization for the inference MLP for deviations.
+        Either None, 'batch' or 'layer'.
+    dev_inf_activation : str
+        Type of activation to use for the inference MLP for deviations.
+        E.g. 'relu', 'sigmoid', 'tanh', etc.
+    dev_inf_dropout : float
+        Dropout rate for the inference MLP for deviations.
+    dev_mod_n_layers : int
+        Number of layers for the model MLP for deviations.
+    dev_mod_n_units : int
+        Number of units for the model MLP for deviations.
+    dev_mod_normalization : str
+        Type of normalization for the model MLP for deviations.
+    dev_mod_activation : str
+        Type of activation to use for the model MLP for deviations.
+    dev_mod_dropout : float
+        Dropout rate for the model MLP for deviations.
 
     batch_size : int
         Mini-batch size.
@@ -102,6 +114,7 @@ class Config(BaseModelConfig):
     learn_covariances: bool = None
     initial_means: np.ndarray = None
     initial_covariances: np.ndarray = None
+    covariances_epsilon: float = None
     means_regularizer: tf.keras.regularizers.Regularizer = None
     covariances_regularizer: tf.keras.regularizers.Regularizer = None
 
@@ -109,13 +122,17 @@ class Config(BaseModelConfig):
     n_subjects: int = None
     subject_embedding_dim: int = None
     mode_embedding_dim: int = None
-    mode_embedding_activation: str = None
-    dev_reg: str = None
-    dev_reg_strength: float = 0.0
-    dev_bayesian: bool = False
-    learn_dev_mod_sigma: bool = True
-    initial_dev_mod_sigma: float = 1
-    do_kl_annealing: bool = False
+
+    dev_inf_n_layers: int = 0
+    dev_inf_n_units: int = None
+    dev_inf_normalization: str = None
+    dev_inf_activation: str = None
+    dev_inf_dropout: float = 0.0
+    dev_mod_n_layers: int = 0
+    dev_mod_n_units: int = None
+    dev_mod_normalization: str = None
+    dev_mod_activation: str = None
+    dev_mod_dropout: float = 0.0
 
     def __post_init__(self):
         self.validate_observation_model_parameters()
@@ -127,6 +144,12 @@ class Config(BaseModelConfig):
         if self.learn_means is None or self.learn_covariances is None:
             raise ValueError("learn_means and learn_covariances must be passed.")
 
+        if self.covariances_epsilon is None:
+            if self.learn_covariances:
+                self.covariances_epsilon = 1e-6
+            else:
+                self.covariances_epsilon = 0.0
+
     def validate_subject_embedding_parameters(self):
         if (
             self.n_subjects is None
@@ -137,6 +160,12 @@ class Config(BaseModelConfig):
                 "n_subjects, subject_embedding_dim and mode_embedding_dim must be passed."
             )
 
+        if self.dev_inf_n_layers != 0 and self.dev_inf_n_units is None:
+            raise ValueError("Please pass dev_inf_n_units.")
+
+        if self.dev_mod_n_layers != 0 and self.dev_mod_n_units is None:
+            raise ValueError("Plase pass dev_mod_n_units.")
+
 
 class Model(ModelBase):
     """SE-DyNeMo observation model class.
@@ -211,7 +240,7 @@ class Model(ModelBase):
             Deviation of Cholesky factor of covs from the group.
             Shaoe is (n_subjects, n_modes, n_channels * (n_channels + 1) // 2).
         """
-        return get_subject_dev(self.model, self.config.dev_bayesian)
+        return get_subject_dev(self.model)
 
     def get_subject_means_covariances(self):
         """Get the means and covariances for each subject
@@ -223,7 +252,7 @@ class Model(ModelBase):
         subject_covs : np.ndarray
             Mode covariances for each subject. Shape is (n_subjects, n_modes, n_channels, n_channels).
         """
-        return get_subject_means_covariances(self.model, self.config.dev_bayesian)
+        return get_subject_means_covariances(self.model)
 
     def set_regularizers(self, training_dataset):
         """Set the means and covariances regularizer based on the training data.
@@ -319,17 +348,16 @@ def _model_structure(config):
         config.n_channels,
         config.learn_covariances,
         config.initial_covariances,
+        config.covariances_epsilon,
         config.covariances_regularizer,
         name="group_covs",
     )
     means_mode_embedding_layer = layers.Dense(
         config.mode_embedding_dim,
-        config.mode_embedding_activation,
         name="means_mode_embedding",
     )
     covs_mode_embedding_layer = layers.Dense(
         config.mode_embedding_dim,
-        config.mode_embedding_activation,
         name="covs_mode_embedding",
     )
     means_concat_embedding_layer = SubjectDevEmbeddingLayer(
@@ -344,87 +372,70 @@ def _model_structure(config):
         config.n_subjects,
         name="covs_concat_embedding",
     )
-    # ----------------------------------------- #
-    # Layers specific to the non Bayesian model #
-    # ----------------------------------------- #
-    if not config.dev_bayesian:
-        if config.learn_means:
-            means_dev_layer = layers.Dense(config.n_channels, name="means_dev")
-        else:
-            means_dev_layer = ZeroLayer(
-                shape=(config.n_subjects, config.n_modes, config.n_channels),
-                name="means_dev",
-            )
 
-        if config.learn_covariances:
-            covs_dev_layer = layers.Dense(
-                config.n_channels * (config.n_channels + 1) // 2,
-                name="covs_dev",
-            )
-        else:
-            covs_dev_layer = ZeroLayer(
-                shape=(
-                    config.n_subjects,
-                    config.n_modes,
-                    config.n_channels * (config.n_channels + 1) // 2,
-                ),
-                name="covs_dev",
-            )
-
-        means_dev_reg_layer = DummyLayer(
-            config.dev_reg,
-            config.dev_reg_strength,
-            name="means_dev_reg",
+    # Inference part of the deviation
+    means_dev_inf_input_layer = MLPLayer(
+        config.dev_inf_n_layers,
+        config.dev_inf_n_units,
+        config.dev_inf_normalization,
+        config.dev_inf_activation,
+        config.dev_inf_dropout,
+        name="means_dev_inf_input",
+    )
+    means_dev_inf_mu_layer = layers.Dense(config.n_channels, name="means_dev_inf_mu")
+    means_dev_inf_sigma_layer = layers.Dense(
+        config.n_channels, activation="softplus", name="means_dev_inf_sigma"
+    )
+    if config.learn_means:
+        means_dev_layer = SampleNormalDistributionLayer(
+            config.theta_std_epsilon, name="means_dev"
         )
-        covs_dev_reg_layer = DummyLayer(
-            config.dev_reg,
-            config.dev_reg_strength,
-            name="covs_dev_reg",
-        )
-    # ------------------------------------- #
-    # Layers specific to the Bayesian model #
-    # ------------------------------------- #
     else:
-        means_dev_inf_mu_layer = layers.Dense(
-            config.n_channels, name="means_dev_inf_mu"
+        means_dev_layer = ZeroLayer(
+            shape=(config.n_subjects, config.n_modes, config.n_channels),
+            name="means_dev",
         )
-        means_dev_inf_sigma_layer = layers.Dense(
-            config.n_channels, activation="softplus", name="means_dev_inf_sigma"
-        )
-        if config.learn_means:
-            means_dev_layer = SampleNormalDistributionLayer(name="means_dev")
-        else:
-            means_dev_layer = ZeroLayer(
-                shape=(config.n_subjects, config.n_modes, config.n_channels),
-                name="means_dev",
-            )
 
-        covs_dev_inf_mu_layer = layers.Dense(
-            config.n_channels * (config.n_channels + 1) // 2, name="covs_dev_inf_mu"
+    covs_dev_inf_input_layer = MLPLayer(
+        config.dev_inf_n_layers,
+        config.dev_inf_n_units,
+        config.dev_inf_normalization,
+        config.dev_inf_activation,
+        config.dev_inf_dropout,
+        name="covs_dev_inf_input",
+    )
+    covs_dev_inf_mu_layer = layers.Dense(
+        config.n_channels * (config.n_channels + 1) // 2, name="covs_dev_inf_mu"
+    )
+    covs_dev_inf_sigma_layer = layers.Dense(
+        config.n_channels * (config.n_channels + 1) // 2,
+        activation="softplus",
+        name="covs_dev_inf_sigma",
+    )
+    if config.learn_covariances:
+        covs_dev_layer = SampleNormalDistributionLayer(
+            config.theta_std_epsilon, name="covs_dev"
         )
-        covs_dev_inf_sigma_layer = layers.Dense(
-            config.n_channels * (config.n_channels + 1) // 2,
-            activation="softplus",
-            name="covs_dev_inf_sigma",
+    else:
+        covs_dev_layer = ZeroLayer(
+            shape=(
+                config.n_subjects,
+                config.n_modes,
+                config.n_channels * (config.n_channels + 1) // 2,
+            ),
+            name="covs_dev",
         )
-        if config.learn_covariances:
-            covs_dev_layer = SampleNormalDistributionLayer(name="covs_dev")
-        else:
-            covs_dev_layer = ZeroLayer(
-                shape=(
-                    config.n_subjects,
-                    config.n_modes,
-                    config.n_channels * (config.n_channels + 1) // 2,
-                ),
-                name="covs_dev",
-            )
 
-    subject_means_layer = SubjectMapLayer("means", name="subject_means")
-    subject_covs_layer = SubjectMapLayer("covariances", name="subject_covs")
+    subject_means_layer = SubjectMapLayer(
+        "means", config.covariances_epsilon, name="subject_means"
+    )
+    subject_covs_layer = SubjectMapLayer(
+        "covariances", config.covariances_epsilon, name="subject_covs"
+    )
     mix_subject_means_covs_layer = MixSubjectEmbeddingParametersLayer(
         name="mix_subject_means_covs"
     )
-    ll_loss_layer = LogLikelihoodLossLayer(name="ll_loss")
+    ll_loss_layer = LogLikelihoodLossLayer(config.covariances_epsilon, name="ll_loss")
 
     # Data flow
     subjects = subjects_layer(data)  # data not used here
@@ -435,7 +446,9 @@ def _model_structure(config):
 
     # spatial map embeddings
     means_mode_embedding = means_mode_embedding_layer(group_mu)
-    covs_mode_embedding = covs_mode_embedding_layer(InverseCholeskyLayer()(group_D))
+    covs_mode_embedding = covs_mode_embedding_layer(
+        InverseCholeskyLayer(config.covariances_epsilon)(group_D)
+    )
 
     # Now get the subject specific spatial maps
     means_concat_embedding = means_concat_embedding_layer(
@@ -445,20 +458,15 @@ def _model_structure(config):
         [subject_embeddings, covs_mode_embedding]
     )
 
-    if not config.dev_bayesian:
-        means_dev = means_dev_layer(means_concat_embedding)
-        means_dev = means_dev_reg_layer(means_dev)
+    means_dev_inf_input = means_dev_inf_input_layer(means_concat_embedding)
+    means_dev_inf_mu = means_dev_inf_mu_layer(means_dev_inf_input)
+    means_dev_inf_sigma = means_dev_inf_sigma_layer(means_dev_inf_input)
+    means_dev = means_dev_layer([means_dev_inf_mu, means_dev_inf_sigma])
 
-        covs_dev = covs_dev_layer(covs_concat_embedding)
-        covs_dev = covs_dev_reg_layer(covs_dev)
-    else:
-        means_dev_inf_mu = means_dev_inf_mu_layer(means_concat_embedding)
-        means_dev_inf_sigma = means_dev_inf_sigma_layer(means_concat_embedding)
-        means_dev = means_dev_layer([means_dev_inf_mu, means_dev_inf_sigma])
-
-        covs_dev_inf_mu = covs_dev_inf_mu_layer(covs_concat_embedding)
-        covs_dev_inf_sigma = covs_dev_inf_sigma_layer(covs_concat_embedding)
-        covs_dev = covs_dev_layer([covs_dev_inf_mu, covs_dev_inf_sigma])
+    covs_dev_inf_input = covs_dev_inf_input_layer(covs_concat_embedding)
+    covs_dev_inf_mu = covs_dev_inf_mu_layer(covs_dev_inf_input)
+    covs_dev_inf_sigma = covs_dev_inf_sigma_layer(covs_dev_inf_input)
+    covs_dev = covs_dev_layer([covs_dev_inf_mu, covs_dev_inf_sigma])
 
     mu = subject_means_layer([group_mu, means_dev])
     D = subject_covs_layer([group_D, covs_dev])
@@ -467,59 +475,80 @@ def _model_structure(config):
     m, C = mix_subject_means_covs_layer([alpha, mu, D, subj_id])
     ll_loss = ll_loss_layer([data, m, C])
 
-    if not config.dev_bayesian:
-        return tf.keras.Model(
-            inputs=[data, alpha, subj_id], outputs=[ll_loss], name="Se-DyNeMo-Obs"
+    # Prior part of deviations
+    means_dev_mod_input_layer = MLPLayer(
+        config.dev_inf_n_layers,
+        config.dev_inf_n_units,
+        config.dev_inf_normalization,
+        config.dev_inf_activation,
+        config.dev_inf_dropout,
+        name="means_dev_mod_input",
+    )
+    means_dev_mod_sigma_layer = layers.Dense(
+        config.n_channels, activation="softplus", name="means_dev_mod_sigma"
+    )
+    covs_dev_mod_input_layer = MLPLayer(
+        config.dev_inf_n_layers,
+        config.dev_inf_n_units,
+        config.dev_inf_normalization,
+        config.dev_inf_activation,
+        config.dev_inf_dropout,
+        name="covs_dev_mod_input",
+    )
+    covs_dev_mod_sigma_layer = layers.Dense(
+        config.n_channels * (config.n_channels + 1) // 2,
+        activation="softplus",
+        name="covs_dev_mod_sigma",
+    )
+
+    if config.learn_means:
+        means_dev_kl_loss_layer = SubjectMapKLDivergenceLayer(
+            config.theta_std_epsilon, name="means_dev_kl_loss"
         )
     else:
-        # Layers for dev prior
-        dev_mod_sigma_layer = ScalarLayer(
-            config.learn_dev_mod_sigma,
-            config.initial_dev_mod_sigma,
-            name="dev_mod_sigma",
-        )
-        if config.learn_means:
-            means_dev_kl_loss_layer = SubjectMapKLDivergenceLayer(
-                name="means_dev_kl_loss"
-            )
-        else:
-            means_dev_kl_loss_layer = ZeroLayer((), name="means_dev_kl_loss")
+        means_dev_kl_loss_layer = ZeroLayer((), name="means_dev_kl_loss")
 
-        if config.learn_covariances:
-            covs_dev_kl_loss_layer = SubjectMapKLDivergenceLayer(
-                name="covs_dev_kl_loss"
-            )
-        else:
-            covs_dev_kl_loss_layer = ZeroLayer((), name="covs_dev_kl_loss")
-
-        kl_loss_layer = KLLossLayer(do_annealing=config.do_kl_annealing, name="kl_loss")
-
-        # Data flow
-        dev_mod_sigma = dev_mod_sigma_layer(data)  # Data not used
-        means_dev_kl_loss = means_dev_kl_loss_layer(
-            [data, means_dev_inf_mu, means_dev_inf_sigma, dev_mod_sigma]
+    if config.learn_covariances:
+        covs_dev_kl_loss_layer = SubjectMapKLDivergenceLayer(
+            config.theta_std_epsilon, name="covs_dev_kl_loss"
         )
-        covs_dev_kl_loss = covs_dev_kl_loss_layer(
-            [data, covs_dev_inf_mu, covs_dev_inf_sigma, dev_mod_sigma]
-        )
-        kl_loss = kl_loss_layer([means_dev_kl_loss, covs_dev_kl_loss])
+    else:
+        covs_dev_kl_loss_layer = ZeroLayer((), name="covs_dev_kl_loss")
 
-        return tf.keras.Model(
-            inputs=[data, alpha, subj_id],
-            outputs=[ll_loss, kl_loss],
-            name="Se_DyNeMo-Obs",
-        )
+    kl_loss_layer = KLLossLayer(do_annealing=config.do_kl_annealing, name="kl_loss")
+
+    # Data flow
+    means_dev_mod_input = means_dev_mod_input_layer(means_concat_embedding)
+    means_dev_mod_sigma = means_dev_mod_sigma_layer(means_dev_mod_input)
+
+    covs_dev_mod_input = covs_dev_mod_input_layer(covs_concat_embedding)
+    covs_dev_mod_sigma = covs_dev_mod_sigma_layer(covs_dev_mod_input)
+
+    means_dev_kl_loss = means_dev_kl_loss_layer(
+        [data, means_dev_inf_mu, means_dev_inf_sigma, means_dev_mod_sigma]
+    )
+    covs_dev_kl_loss = covs_dev_kl_loss_layer(
+        [data, covs_dev_inf_mu, covs_dev_inf_sigma, covs_dev_mod_sigma]
+    )
+    kl_loss = kl_loss_layer([means_dev_kl_loss, covs_dev_kl_loss])
+
+    return tf.keras.Model(
+        inputs=[data, alpha, subj_id],
+        outputs=[ll_loss, kl_loss],
+        name="Se_DyNeMo-Obs",
+    )
 
 
 def get_group_means_covariances(model):
     group_means_layer = model.get_layer("group_means")
     group_covs_layer = model.get_layer("group_covs")
 
-    group_means = group_means_layer.vectors.numpy()
-    group_covs = group_covs_layer.bijector(
-        group_covs_layer.flattened_cholesky_factors
-    ).numpy()
-    return group_means, group_covs
+    group_means = group_means_layer.vectors
+    group_covs = add_epsilon(
+        group_covs_layer.bijector(group_covs_layer.flattened_cholesky_factors),
+        group_covs_layer.epsilon,
+    )
+    return group_means.numpy(), group_covs.numpy()
 
 
 def get_subject_embeddings(model):
@@ -559,28 +588,24 @@ def get_concatenated_embeddings(model):
     return means_concat_embedding.numpy(), covs_concat_embedding.numpy()
 
 
-def get_subject_dev(model, dev_bayesian):
+def get_subject_dev(model):
     means_concat_embedding, covs_concat_embedding = get_concatenated_embeddings(model)
-    means_dev_layer = (
-        model.get_layer("means_dev_inf_mu")
-        if dev_bayesian
-        else model.get_layer("means_dev")
-    )
-    covs_dev_layer = (
-        model.get_layer("covs_dev_inf_mu")
-        if dev_bayesian
-        else model.get_layer("covs_dev")
-    )
+    means_dev_inf_input_layer = model.get_layer("means_dev_inf_input")
+    covs_dev_inf_input_layer = model.get_layer("covs_dev_inf_input")
+    means_dev_inf_layer = model.get_layer("means_dev_inf_mu")
+    covs_dev_inf_layer = model.get_layer("covs_dev_inf_mu")
 
-    means_dev = means_dev_layer(means_concat_embedding)
-    covs_dev = covs_dev_layer(covs_concat_embedding)
+    means_dev_inf_input = means_dev_inf_input_layer(means_concat_embedding)
+    covs_dev_inf_input = covs_dev_inf_input_layer(covs_concat_embedding)
+    means_dev = means_dev_inf_layer(means_dev_inf_input)
+    covs_dev = covs_dev_inf_layer(covs_dev_inf_input)
 
     return means_dev.numpy(), covs_dev.numpy()
 
 
-def get_subject_means_covariances(model, dev_bayesian):
+def get_subject_means_covariances(model):
     group_means, group_covs = get_group_means_covariances(model)
-    means_dev, covs_dev = get_subject_dev(model, dev_bayesian)
+    means_dev, covs_dev = get_subject_dev(model)
 
     subject_means_layer = model.get_layer("subject_means")
     subject_covs_layer = model.get_layer("subject_covs")
