@@ -134,17 +134,17 @@ def rough_square_axes(n_plots):
     return short, long, empty
 
 
-def get_colors(n_modes, colormap="magma"):
+def get_colors(n, colormap="magma"):
     """Produce equidistant colors from a matplotlib colormap.
 
     Given a matplotlib colormap, produce a series of RGBA colors which are equally
     spaced by value. There is no guarantee that these colors will be perceptually
-    uniformly distributed and with many colors will likely be extremely close. Alpha is
-    1.0 for all colors.
+    uniformly distributed and with many colors will likely be extremely close.
+    Alpha is 1.0 for all colors.
 
     Parameters
     ----------
-    n_modes : int
+    n : int
         The number of colors to return.
     colormap : str
         The name of a matplotlib colormap.
@@ -155,7 +155,7 @@ def get_colors(n_modes, colormap="magma"):
         A list of colors in RGBA format. A = 1.0 in all cases.
     """
     colormap = plt.get_cmap(colormap)
-    colors = [colormap(1 * i / n_modes) for i in range(n_modes)]
+    colors = [colormap(1 * i / n) for i in range(n)]
     return colors
 
 
@@ -169,8 +169,6 @@ def plot_line(
     y_range=None,
     x_label=None,
     y_label=None,
-    log_xscale=False,
-    log_yscale=False,
     title=None,
     plot_kwargs=None,
     fig_kwargs=None,
@@ -419,7 +417,10 @@ def plot_scatter(
     if plot_kwargs is None:
         plot_kwargs = {}
 
-    colors = get_colors(len(x), colormap="tab10")
+    if len(x) > 10:
+        colors = get_colors(len(x), colormap="tab20")
+    else:
+        colors = get_colors(len(x), colormap="tab10")
 
     # Create figure
     if ax is None:
@@ -689,7 +690,7 @@ def plot_gmm(
     data,
     amplitudes,
     means,
-    variances,
+    stddevs,
     bins=50,
     legend_loc=1,
     x_range=None,
@@ -710,8 +711,8 @@ def plot_gmm(
         Raw data to plot as a histogram.
     means : np.ndarray
         Mean of each Gaussian component.
-    variances : np.ndarray
-        Variance of each Gaussian component.
+    stddevs : np.ndarray
+        Standard deviation of each Gaussian component.
     bins : list of int
         Number of bins for the historgram.
     legend_loc : int
@@ -777,8 +778,8 @@ def plot_gmm(
 
     # Plot Gaussian components
     x = np.arange(min(data), max(data), (max(data) - min(data)) / bins)
-    y1 = amplitudes[0] * np.exp(-((x - means[0]) ** 2) / (2 * variances[0] ** 2))
-    y2 = amplitudes[1] * np.exp(-((x - means[1]) ** 2) / (2 * variances[1] ** 2))
+    y1 = amplitudes[0] * np.exp(-((x - means[0]) ** 2) / (2 * stddevs[0] ** 2))
+    y2 = amplitudes[1] * np.exp(-((x - means[1]) ** 2) / (2 * stddevs[1] ** 2))
     ax.plot(x, y1, label="Off")
     ax.plot(x, y2, label="On")
     ax.plot(x, y1 + y2)
